"""List-archive command implementation - list contents of archive files."""

from __future__ import annotations

import pathlib

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from argparse import Namespace
    from logging import Logger

# Pykotorcli-level filter supports matching either "resref" or "resref.ext".
from pykotor.cli.archive_filter import matches_resource_name
from pykotor.tools.archives import list_bif, list_erf, list_key, list_rim

# vendor references:


def cmd_list_archive(args: Namespace, logger: Logger) -> int:  # noqa: PLR0911
    """Handle list-archive command - list contents of archive files.

    Supports:
    - KEY/BIF archives
    - RIM archives
    - ERF/MOD/SAV/HAK archives

    Args:
    ----
        args: Parsed command line arguments
        logger: Logger instance

    Returns:
    -------
        Exit code (0 for success, non-zero for error)
    """
    if not args.file:
        logger.error("No input file specified. Use --file <archive>")
        return 1

    input_path = pathlib.Path(args.file)
    if not input_path.exists():
        logger.error(f"Input file not found: {input_path}")  # noqa: G004
        return 1

    suffix = input_path.suffix.lower()
    logger.info(f"Listing contents of {suffix} archive: {input_path.name}")  # noqa: G004

    try:
        if suffix == ".key":
            return _list_key(input_path, args, logger)
        if suffix == ".bif":
            return _list_bif(input_path, args, logger)
        if suffix in (".rim",):
            return _list_rim(input_path, args, logger)
        if suffix in (".erf", ".mod", ".sav", ".hak"):
            return _list_erf(input_path, args, logger)
        logger.error(f"Unsupported archive type: {suffix}")  # noqa: G004
        logger.info("Supported types: .key, .bif, .rim, .erf, .mod, .sav, .hak")
    except Exception:
        logger.exception("Failed to list archive")
        return 1
    else:
        return 0


def _list_key(key_path: pathlib.Path, args: Namespace, logger: Logger) -> int:
    """List contents of KEY file and associated BIF files."""
    try:
        bif_files, resources = list_key(key_path)

        logger.info(f"KEY file: {key_path.name}")  # noqa: G004
        logger.info(f"  BIF files: {len(bif_files)}")  # noqa: G004
        logger.info(f"  Resources: {len(resources)}")  # noqa: G004

        if args.verbose:
            logger.info("\nBIF Files:")
            for i, bif_name in enumerate(bif_files):
                logger.info(f"  [{i}] {bif_name}")  # noqa: G004

        if args.resources or not args.bifs_only:
            logger.info("\nResources:")
            for resref, res_type, bif_index, res_index in resources:
                if args.filter and not matches_resource_name(resref, res_type, args.filter):
                    continue

                bif_name = bif_files[bif_index] if bif_index < len(bif_files) else "?"
                logger.info(f"  {resref}.{res_type} (BIF: {bif_name}, idx: {res_index})")  # noqa: G004

    except Exception:
        logger.exception("Failed to list KEY")
        return 1
    else:
        return 0


def _list_bif(bif_path: pathlib.Path, args: Namespace, logger: Logger) -> int:
    """List contents of a BIF file."""
    try:
        key_path = pathlib.Path(args.key_file) if args.key_file else bif_path.parent / "chitin.key"
        resources = list(list_bif(bif_path, key_path=key_path if key_path.exists() else None))

        logger.info(f"BIF file: {bif_path.name}")  # noqa: G004
        logger.info(f"  Resources: {len(resources)}")  # noqa: G004

        # Backward-compat: historically, listing individual resources required --verbose.
        # We also allow explicit opt-in via --resources, or when a --filter is provided.
        should_print_resources = bool(args.verbose or args.resources or args.filter)
        if should_print_resources:
            logger.info("\nResources:")
            for i, resource in enumerate(resources):
                resref = resource.resref.get() if resource.resref else f"resource_{i:05d}"
                res_type = resource.restype.extension if resource.restype else "bin"
                if args.filter and not matches_resource_name(resref, res_type, args.filter):
                    continue

                if args.verbose:
                    size = len(resource.data) if hasattr(resource, "data") else 0
                    logger.info(f"  [{i}] {resref}.{res_type} ({size} bytes)")  # noqa: G004
                else:
                    logger.info(f"  {resref}.{res_type}")  # noqa: G004

    except Exception:
        logger.exception("Failed to list BIF")
        return 1
    else:
        return 0


def _list_rim(rim_path: pathlib.Path, args: Namespace, logger: Logger) -> int:
    """List contents of a RIM archive."""
    try:
        resources = list(list_rim(rim_path))

        logger.info(f"RIM file: {rim_path.name}")  # noqa: G004
        logger.info(f"  Resources: {len(resources)}")  # noqa: G004

        should_print_resources = bool(args.verbose or args.resources or args.filter)
        if should_print_resources:
            logger.info("\nResources:")
            for resource in resources:
                resref = resource.resref.get() if resource.resref else "unknown"
                res_type = resource.restype.extension if resource.restype else "bin"
                if args.filter and not matches_resource_name(resref, res_type, args.filter):
                    continue

                if args.verbose:
                    size = len(resource.data) if hasattr(resource, "data") else 0
                    logger.info(f"  {resref}.{res_type} ({size} bytes)")  # noqa: G004
                else:
                    logger.info(f"  {resref}.{res_type}")  # noqa: G004

    except Exception:
        logger.exception("Failed to list RIM")
        return 1
    else:
        return 0


def _list_erf(erf_path: pathlib.Path, args: Namespace, logger: Logger) -> int:
    """List contents of an ERF/MOD/SAV/HAK archive."""
    try:
        resources = list(list_erf(erf_path))

        logger.info(f"ERF file: {erf_path.name}")  # noqa: G004
        logger.info(f"  Resources: {len(resources)}")  # noqa: G004

        should_print_resources = bool(args.verbose or args.resources or args.filter)
        if should_print_resources:
            logger.info("\nResources:")
            for resource in resources:
                resref = resource.resref.get() if resource.resref else "unknown"
                res_type = resource.restype.extension if resource.restype else "bin"
                if args.filter and not matches_resource_name(resref, res_type, args.filter):
                    continue

                if args.verbose:
                    size = len(resource.data) if hasattr(resource, "data") else 0
                    logger.info(f"  {resref}.{res_type} ({size} bytes)")  # noqa: G004
                else:
                    logger.info(f"  {resref}.{res_type}")  # noqa: G004

    except Exception:
        logger.exception("Failed to list ERF")
        return 1
    else:
        return 0



from __future__ import annotations

import ctypes
import ctypes.wintypes

from enum import IntEnum, IntFlag
from typing import Sequence


class JOBOBJECTINFOCLASS(IntEnum):
    """https://learn.microsoft.com/en-us/windows/win32/api/winnt/ne-winnt-jobobject_limit_information_class."""

    JobObjectBasicAccountingInformation = 1
    JobObjectBasicLimitInformation = 2
    JobObjectBasicProcessIdList = 3
    JobObjectBasicUIRestrictions = 4
    JobObjectSecurityLimitInformation = 5
    JobObjectEndOfJobTimeInformation = 6
    JobObjectAssociateCompletionPortInformation = 7
    JobObjectBasicAndIoAccountingInformation = 8
    JobObjectExtendedLimitInformation = 9
    JobObjectJobSetInformation = 10
    JobObjectGroupInformation = 11
    JobObjectNotificationLimitInformation = 12
    JobObjectLimitViolationInformation = 13
    JobObjectGroupInformationEx = 14
    JobObjectCpuRateControlInformation = 15
    JobObjectCompletionFilter = 16
    JobObjectCompletionCounter = 17
    JobObjectReserved1Information = 18
    JobObjectReserved2Information = 19
    JobObjectReserved3Information = 20
    JobObjectReserved4Information = 21
    JobObjectReserved5Information = 22
    JobObjectReserved6Information = 23
    JobObjectReserved7Information = 24
    JobObjectReserved8Information = 25
    MaxJobObjectInfoClass = 26


class JOBOBJECT_BASIC_LIMIT_INFORMATION(ctypes.Structure):  # noqa: N801
    _fields_: Sequence[tuple[str, type[ctypes._CData]] | tuple[str, type[ctypes._CData], int]] = [
        ("PerProcessUserTimeLimit", ctypes.wintypes.LARGE_INTEGER),
        ("PerJobUserTimeLimit", ctypes.wintypes.LARGE_INTEGER),
        ("LimitFlags", ctypes.wintypes.DWORD),
        ("MinimumWorkingSetSize", ctypes.c_size_t),
        ("MaximumWorkingSetSize", ctypes.c_size_t),
        ("ActiveProcessLimit", ctypes.wintypes.DWORD),
        ("Affinity", ctypes.POINTER(ctypes.c_ulonglong)),
        ("PriorityClass", ctypes.wintypes.DWORD),
        ("SchedulingClass", ctypes.wintypes.DWORD),
    ]
    PerProcessUserTimeLimit: ctypes.wintypes.LARGE_INTEGER  # The time limit for the user time of a process in the job.
    PerJobUserTimeLimit: ctypes.wintypes.LARGE_INTEGER  # The time limit for the user time of the job.
    LimitFlags: ctypes.wintypes.DWORD  # The limit flags for the job object.
    MinimumWorkingSetSize: ctypes.wintypes.LPVOID  # The minimum working set size for the job object.
    MaximumWorkingSetSize: ctypes.wintypes.LPVOID  # The maximum working set size for the job object.
    ActiveProcessLimit: ctypes.wintypes.DWORD  # The maximum number of processes that can be active in the job object.
    Affinity: ctypes._Pointer[ctypes.wintypes.LPVOID]  # The affinity mask for the job object.
    PriorityClass: ctypes.wintypes.DWORD  # The priority class for the job object.
    SchedulingClass: ctypes.wintypes.DWORD  # The scheduling class for the job object.


class IO_COUNTERS(ctypes.Structure):  # noqa: N801
    """https://learn.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-jobobject_io_counters."""

    _fields_: Sequence[tuple[str, type[ctypes._CData]] | tuple[str, type[ctypes._CData], int]] = [
        ("ReadOperationCount", ctypes.wintypes.ULARGE_INTEGER),
        ("WriteOperationCount", ctypes.wintypes.ULARGE_INTEGER),
        ("OtherOperationCount", ctypes.wintypes.ULARGE_INTEGER),
        ("ReadTransferCount", ctypes.wintypes.ULARGE_INTEGER),
        ("WriteTransferCount", ctypes.wintypes.ULARGE_INTEGER),
        ("OtherTransferCount", ctypes.wintypes.ULARGE_INTEGER),
    ]
    ReadOperationCount: ctypes.wintypes.ULARGE_INTEGER  # The number of read operations performed by the job object.
    WriteOperationCount: ctypes.wintypes.ULARGE_INTEGER  # The number of write operations performed by the job object.
    OtherOperationCount: ctypes.wintypes.ULARGE_INTEGER  # The number of other operations performed by the job object.
    ReadTransferCount: ctypes.wintypes.ULARGE_INTEGER  # The number of bytes read from the job object.
    WriteTransferCount: ctypes.wintypes.ULARGE_INTEGER  # The number of bytes written to the job object.
    OtherTransferCount: ctypes.wintypes.ULARGE_INTEGER  # The number of bytes transferred to the job object.


class JOBOBJECT_EXTENDED_LIMIT_INFORMATION(ctypes.Structure):  # noqa: N801
    """https://learn.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-jobobject_extended_limit_information."""

    _fields_: Sequence[tuple[str, type[ctypes._CData]] | tuple[str, type[ctypes._CData], int]] = [
        ("BasicLimitInformation", JOBOBJECT_BASIC_LIMIT_INFORMATION),
        ("IoInfo", IO_COUNTERS),
        ("ProcessMemoryLimit", ctypes.c_size_t),
        ("JobMemoryLimit", ctypes.c_size_t),
        ("PeakProcessMemoryUsed", ctypes.c_size_t),
        ("PeakJobMemoryUsed", ctypes.c_size_t),
    ]
    BasicLimitInformation: JOBOBJECT_BASIC_LIMIT_INFORMATION  # The basic limit information for the job object.
    IoInfo: IO_COUNTERS  # The I/O counters for the job object.
    ProcessMemoryLimit: ctypes.wintypes.LPVOID  # The memory limit for the job object.
    JobMemoryLimit: ctypes.wintypes.LPVOID  # The memory limit for the job object.
    PeakProcessMemoryUsed: ctypes.wintypes.LPVOID  # The peak memory used by the job object.
    PeakJobMemoryUsed: ctypes.wintypes.LPVOID  # The peak memory used by the job object.


class JOBOBJECT_BASIC_PROCESS_ID_LIST(ctypes.Structure):  # noqa: N801
    _fields_: Sequence[tuple[str, type[ctypes._CData]] | tuple[str, type[ctypes._CData], int]] = [
        ("NumberOfAssignedProcesses", ctypes.wintypes.DWORD),
        ("NumberOfProcessIdsInList", ctypes.wintypes.DWORD),
        ("ProcessIdList", ctypes.POINTER(ctypes.wintypes.DWORD)),
    ]
    NumberOfAssignedProcesses: ctypes.wintypes.DWORD  # The number of processes assigned to the job object.
    NumberOfProcessIdsInList: ctypes.wintypes.DWORD  # The number of process IDs in the list.
    ProcessIdList: ctypes._Pointer[ctypes.wintypes.DWORD]  # The list of process IDs assigned to the job object.


class JOBOBJECT_BASIC_ACCOUNTING_INFORMATION(ctypes.Structure):  # noqa: N801
    _fields_: Sequence[tuple[str, type[ctypes._CData]] | tuple[str, type[ctypes._CData], int]] = [
        ("TotalUserTime", ctypes.wintypes.LARGE_INTEGER),
        ("TotalKernelTime", ctypes.wintypes.LARGE_INTEGER),
        ("ThisPeriodTotalUserTime", ctypes.wintypes.LARGE_INTEGER),
        ("ThisPeriodTotalKernelTime", ctypes.wintypes.LARGE_INTEGER),
        ("TotalPageFaultCount", ctypes.wintypes.DWORD),
        ("TotalProcesses", ctypes.wintypes.DWORD),
        ("ActiveProcesses", ctypes.wintypes.DWORD),
        ("TotalTerminatedProcesses", ctypes.wintypes.DWORD),
    ]
    TotalUserTime: ctypes.wintypes.LARGE_INTEGER  # The total user time for the job object.
    TotalKernelTime: ctypes.wintypes.LARGE_INTEGER  # The total kernel time for the job object.
    ThisPeriodTotalUserTime: ctypes.wintypes.LARGE_INTEGER  # The total user time for the current period for the job object.
    ThisPeriodTotalKernelTime: ctypes.wintypes.LARGE_INTEGER  # The total kernel time for the current period for the job object.
    TotalPageFaultCount: ctypes.wintypes.DWORD  # The total number of page faults for the job object.
    TotalProcesses: ctypes.wintypes.DWORD  # The total number of processes for the job object.
    ActiveProcesses: ctypes.wintypes.DWORD  # The number of active processes for the job object.
    TotalTerminatedProcesses: ctypes.wintypes.DWORD  # The total number of terminated processes for the job object.


class JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION(ctypes.Structure):  # noqa: N801
    _fields_: Sequence[tuple[str, type[ctypes._CData]] | tuple[str, type[ctypes._CData], int]] = [
        ("BasicInfo", JOBOBJECT_BASIC_ACCOUNTING_INFORMATION),
        ("IoInfo", IO_COUNTERS),
    ]
    BasicInfo: JOBOBJECT_BASIC_ACCOUNTING_INFORMATION  # The basic accounting information for the job object.
    IoInfo: IO_COUNTERS  # The I/O counters for the job object.


class JOBOBJECTLIMIT(IntFlag):
    """https://learn.microsoft.com/en-us/windows/win32/api/winnt/ne-winnt-jobobject_limit_information_class."""

    JOB_OBJECT_LIMIT_WORKINGSET = 0x00000001
    JOB_OBJECT_LIMIT_PROCESS_TIME = 0x00000002
    JOB_OBJECT_LIMIT_JOB_TIME = 0x00000004
    JOB_OBJECT_LIMIT_ACTIVE_PROCESS = 0x00000008
    JOB_OBJECT_LIMIT_AFFINITY = 0x00000010
    JOB_OBJECT_LIMIT_PRIORITY_CLASS = 0x00000020
    JOB_OBJECT_LIMIT_PRESERVE_JOB_TIME = 0x00000040
    JOB_OBJECT_LIMIT_SCHEDULING_CLASS = 0x00000080
    JOB_OBJECT_LIMIT_PROCESS_MEMORY = 0x00000100
    JOB_OBJECT_LIMIT_JOB_MEMORY = 0x00000200
    JOB_OBJECT_LIMIT_DIE_ON_UNHANDLED_EXCEPTION = 0x00000400
    JOB_OBJECT_LIMIT_BREAKAWAY_OK = 0x00000800
    JOB_OBJECT_LIMIT_SILENT_BREAKAWAY_OK = 0x00001000
    JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE = 0x00002000
    JOB_OBJECT_LIMIT_SUBSET_AFFINITY = 0x00004000
    JOB_OBJECT_LIMIT_JOB_MEMORY_LOW = 0x00008000
    JOB_OBJECT_LIMIT_JOB_READ_BYTES = 0x00010000
    JOB_OBJECT_LIMIT_JOB_WRITE_BYTES = 0x00020000
    JOB_OBJECT_LIMIT_RATE_CONTROL = 0x00040000
    JOB_OBJECT_LIMIT_CPU_RATE_CONTROL = 0x00080000
    JOB_OBJECT_LIMIT_IO_RATE_CONTROL = 0x00100000
    JOB_OBJECT_LIMIT_NET_RATE_CONTROL = 0x00200000


def get_job_limits(job: int) -> IO_COUNTERS:
    """Retrieves the I/O counters for the specified job object.

    Further reading:
    - https://learn.microsoft.com/en-us/windows/win32/api/handleapi/nf-handleapi-duplicatehandle
    - https://learn.microsoft.com/en-us/windows/win32/api/handleapi/nf-handleapi-ntclose
    - https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-getcurrentprocess
    - https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-createprocessa
    - https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-openprocess

    Args:
        job (int): The handle to the job object.

    Returns:
        IO_COUNTERS: The I/O counters for the job object.
    """
    counters = IO_COUNTERS()
    info = JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION()
    ret_length = ctypes.wintypes.DWORD()

    if not ctypes.windll.kernel32.QueryInformationJobObject(
        job,
        JOBOBJECTINFOCLASS.JobObjectBasicAndIoAccountingInformation,
        ctypes.byref(info),
        ctypes.sizeof(info),
        ctypes.byref(ret_length),
    ):
        raise ctypes.WinError(ctypes.get_last_error())

    counters = info.IoInfo
    return counters


def close_handle(handle: int) -> None:
    """Closes the specified handle.

    Further reading:
    - https://learn.microsoft.com/en-us/windows/win32/api/handleapi/nf-handleapi-closehandle
    - https://learn.microsoft.com/en-us/windows/win32/api/handleapi/nf-handleapi-duplicatehandle
    - https://learn.microsoft.com/en-us/windows/win32/api/handleapi/nf-handleapi-ntclose
    - https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-getcurrentprocess
    - https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-createprocessa
    - https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-openprocess


    Args:
        handle (int): The handle to close.
    """
    try:
        if not ctypes.windll.kernel32.CloseHandle(handle):
            raise ctypes.WinError(ctypes.get_last_error())  # noqa: TRY301
    except Exception as e:  # noqa: BLE001
        from loggerplus import RobustLogger

        RobustLogger().warning(f"Failed to close handle {handle} using CloseHandle: {e}")

        try:
            ctypes.windll.kernel32.DuplicateHandle(
                ctypes.windll.kernel32.GetCurrentProcess(),  # source process
                handle,  # source handle
                None,  # new process
                0,  # new handle
                False,  # not inheritable  # noqa: FBT003
                2,  # DUPLICATE_CLOSE_SOURCE
            )
        except Exception as e:  # noqa: BLE001
            RobustLogger().warning(f"Failed to close handle {handle} using DuplicateHandle: {e}")

            try:
                # Use NtClose from ntdll.dll as a last resort
                ntdll = ctypes.WinDLL("ntdll.dll")
                ntdll.NtClose.argtypes = [ctypes.wintypes.HANDLE]
                ntdll.NtClose.restype = ctypes.wintypes.LONG
                status = ntdll.NtClose(handle)
                if status < 0:
                    raise ctypes.WinError(ctypes.get_last_error())  # noqa: TRY301
            except Exception as e:
                RobustLogger().error(f"Failed to close handle {handle} using all methods: {e}")
                raise


def create_job_object(
    limit_flags: JOBOBJECTLIMIT | None = JOBOBJECTLIMIT.JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE,
    extended_limit_information: JOBOBJECT_EXTENDED_LIMIT_INFORMATION | None = None,
) -> int:
    """Creates a job object with the specified limit flags and extended limit information.


    Relevant windows documentation:
    - https://learn.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-jobobject_extended_limit_information
    - https://learn.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-jobobject_basic_limit_information
    - https://learn.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-jobobject_io_counters
    - https://learn.microsoft.com/en-us/windows/win32/api/jobapi2/nf-jobapi2-createjobobjecta
    Further reading:
    - https://learn.microsoft.com/en-us/windows/win32/api/jobapi2/nf-jobapi2-assignprocesstojobobject
    - https://learn.microsoft.com/en-us/windows/win32/api/jobapi2/nf-jobapi2-setinformationjobobject
    - https://learn.microsoft.com/en-us/windows/win32/api/jobapi2/nf-jobapi2-getinformationjobobject
    - https://learn.microsoft.com/en-us/windows/win32/api/handleapi/nf-handleapi-closehandle
    - https://learn.microsoft.com/en-us/windows/win32/api/handleapi/nf-handleapi-duplicatehandle

    Args:
        limit_flags (JOBOBJECTLIMIT): The limit flags for the job object.
        extended_limit_information (int): The extended limit information for the job object.

    Returns:
        int: The handle to the job object.
    """
    job = ctypes.windll.kernel32.CreateJobObjectW(None, None)
    if not job:
        raise ctypes.WinError(ctypes.get_last_error())

    if extended_limit_information is None:
        extended_limit_information = JOBOBJECT_EXTENDED_LIMIT_INFORMATION()
    limit_flags_value = limit_flags.value if limit_flags is not None else JOBOBJECTLIMIT.JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE
    extended_limit_information.BasicLimitInformation.LimitFlags = ctypes.c_ulong(limit_flags_value)
    if not ctypes.windll.kernel32.SetInformationJobObject(
        job,
        JOBOBJECTINFOCLASS.JobObjectExtendedLimitInformation,
        ctypes.byref(extended_limit_information),
        ctypes.sizeof(extended_limit_information),
    ):
        raise ctypes.WinError(ctypes.get_last_error())
    return job


def assign_process_to_job_object(job: ctypes.c_int | int, process: ctypes.c_int | int) -> None:
    """Assigns a process to a job object.

    Documentation:
    - https://learn.microsoft.com/en-us/windows/win32/api/jobapi2/nf-jobapi2-assignprocesstojobobject
    - https://learn.microsoft.com/en-us/windows/win32/api/jobapi2/nf-jobapi2-setinformationjobobject
    - https://learn.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-jobobject_extended_limit_information
    - https://learn.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-jobobject_basic_limit_information
    - https://learn.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-jobobject_io_counters
    Further reading:
    - https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-assignprocesstojobobject
    - https://learn.microsoft.com/en-us/windows/win32/api/jobapi2/nf-jobapi2-createjobobjecta
    - https://learn.microsoft.com/en-us/windows/win32/api/jobapi2/nf-jobapi2-setinformationjobobject
    - https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-getcurrentprocess
    - https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-createprocessa
    - https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-openprocess

    Args:
        job (int): The handle to the job object.
        process (int): The handle to the process.
    """
    if not ctypes.windll.kernel32.AssignProcessToJobObject(job, process):
        raise ctypes.WinError(ctypes.get_last_error())  # noqa: TRY301


def set_information_job_object(job: ctypes.c_int | int, info_class: JOBOBJECTINFOCLASS, data: ctypes.Structure) -> None:
    """Sets information for the specified job object.

    Documentation:
    - https://learn.microsoft.com/en-us/windows/win32/api/jobapi2/nf-jobapi2-setinformationjobobject

    Args:
        job (int): The handle to the job object.
        information (int): The information to set.
        data (bytes): The data to set.
    """
    if not ctypes.windll.kernel32.SetInformationJobObject(
        job,
        info_class,
        ctypes.byref(data),
        ctypes.sizeof(data),
    ):
        raise ctypes.WinError(ctypes.get_last_error())

    # Add support for additional information classes
    if info_class == JOBOBJECTINFOCLASS.JobObjectExtendedLimitInformation:
        extended_limit_information = JOBOBJECT_EXTENDED_LIMIT_INFORMATION()
        extended_limit_information.BasicLimitInformation = JOBOBJECT_BASIC_LIMIT_INFORMATION()
        extended_limit_information.BasicLimitInformation.LimitFlags = data.LimitFlags
        extended_limit_information.BasicLimitInformation.MinimumWorkingSetSize = data.MinimumWorkingSetSize
        extended_limit_information.BasicLimitInformation.MaximumWorkingSetSize = data.MaximumWorkingSetSize
        extended_limit_information.BasicLimitInformation.ActiveProcessLimit = data.ActiveProcessLimit
        extended_limit_information.BasicLimitInformation.Affinity = data.Affinity
        extended_limit_information.BasicLimitInformation.PriorityClass = data.PriorityClass
        extended_limit_information.BasicLimitInformation.SchedulingClass = data.SchedulingClass
        if not ctypes.windll.kernel32.SetInformationJobObject(
            job,
            JOBOBJECTINFOCLASS.JobObjectExtendedLimitInformation,
            ctypes.byref(extended_limit_information),
            ctypes.sizeof(extended_limit_information),
        ):
            raise ctypes.WinError(ctypes.get_last_error())


def get_information_job_object(job: ctypes.c_int | int, info_class: JOBOBJECTINFOCLASS, data: ctypes.Structure) -> None:
    ERROR_INSUFFICIENT_BUFFER = 122
    ret_length = ctypes.wintypes.DWORD()
    if not ctypes.windll.kernel32.QueryInformationJobObject(
        job,
        info_class,
        ctypes.byref(data),
        ctypes.sizeof(data),
        ctypes.byref(ret_length),
    ):
        if ctypes.get_last_error() == ERROR_INSUFFICIENT_BUFFER:
            new_data = ctypes.create_string_buffer(ret_length.value)
            if not ctypes.windll.kernel32.QueryInformationJobObject(
                job,
                info_class,
                new_data,
                ret_length.value,
                ctypes.byref(ret_length),
            ):
                raise ctypes.WinError(ctypes.get_last_error())
            ctypes.memmove(ctypes.addressof(data), new_data, ret_length.value)
        else:
            raise ctypes.WinError(ctypes.get_last_error())



from __future__ import annotations

import errno
import os
import platform
import shutil
import sys
import tarfile
import tempfile
import threading
import uuid
import zipfile

from contextlib import suppress
from pathlib import Path, PurePath
from typing import IO, TYPE_CHECKING, Any, Callable

from loggerplus import RobustLogger
from utility.misc import ProcessorArchitecture, ensure_directory_exists, get_normalized_extension
from utility.system.os_helper import (
    get_app_dir,
    get_mac_dot_app_dir,
    is_frozen,
    remove_any,
    win_hide_file,
)
from utility.system.path import ChDir
from utility.updater.downloader import FileDownloader, download_mega_file_url
from utility.updater.restarter import RestartStrategy, Restarter, UpdateStrategy

if TYPE_CHECKING:
    from logging import Logger


class LibUpdate:
    """Used to update library files used by an application. This object is
    returned by pyupdater.client.Client.update_check.

    ######Args:
    """

    def __init__(
        self,
        update_urls: list[str],
        filestem: str,
        current_version: str,
        latest: str,
        progress_hooks: list[Callable[[dict[str, Any]], Any]] | None = None,
        max_download_retries: int | None = None,
        downloader: Callable | None = None,
        http_timeout: int | None = None,
        u_strategy: UpdateStrategy = UpdateStrategy.RENAME,
        r_strategy: RestartStrategy = RestartStrategy.DEFAULT,
        logger: Logger | None = None,
        version_to_tag_parser: Callable | None = None,
    ):
        # If user is using async download this will be True.
        # Future calls to an download methods will not run
        # until the current download is complete. Which will
        # set this back to False.
        self._is_downloading: bool = False
        self.version_to_tag_parser: Callable | None = version_to_tag_parser

        self.update_folder: str = tempfile.mkdtemp("_update", "holotoolset_")
        self.update_temp_path: Path = Path(self.update_folder)

        # Used with the version property.
        # Returns a user friendly version string
        self._version: str = ""

        # List of urls used to look for meta data & updates
        self.update_urls: list[str] = update_urls

        # The name of the asset we are targeting.
        self.filestem: str = filestem

        # The version of the current asset
        self.current_version: str = current_version

        # Progress callbacks
        self.progress_hooks: list[Callable[[dict[str, Any]], Any]] | None = progress_hooks

        # The amount of times to retry a url before giving up
        self.max_download_retries: int | None = max_download_retries

        # HTTP Timeout
        self.http_timeout: int | None = http_timeout

        self.downloader: Callable | None = downloader

        # The update strategy to use
        self.u_strategy: UpdateStrategy = u_strategy

        # The restart strategy to use
        self.r_strategy: RestartStrategy = r_strategy

        self.latest: str = latest

        self.archive_name: str = self.get_archive_names()[0]
        self._current_app_dir: Path = get_app_dir()
        self._download_status: bool = False  # The status of the download. Once downloaded this will be True
        self.log: Logger = logger or RobustLogger()

    def get_expected_filename(self) -> str:
        os_lookup_str: str = platform.system()
        if os_lookup_str == "Windows":
            return f"{self.filestem}.exe"
        if os_lookup_str == "Linux":
            return self.filestem
        if os_lookup_str == "Darwin":
            return f"{self.filestem}.app"
        raise ValueError(f"Unsupported return from platform.system() call: '{os_lookup_str}'")

    @property
    def filename(self) -> str:
        # sourcery skip: assign-if-exp, switch, use-fstring-for-concatenation
        # TODO(th3w1zard1): allow customization of this in the constructor.
        if is_frozen():  # The application is frozen with PyInstaller
            return Path(sys.executable).name
        return self.get_expected_filename()

    @property
    def version(self) -> str:
        """Generates a user friendly version string.

        ######Returns (str): User friendly version string
        """
        if not self._version:
            self._version = self.latest
        channel: dict[int, str] = {0: "Alpha", 1: "Beta"}
        v = list(map(int, self._version.split(".")))

        # 1.2
        version = f"{v[0]}.{v[1]}"
        if v[2] != 0:
            # 1.2.1
            version += f".{v[2]}"
        if v[3] != 2:
            # 1.2.1 Alpha
            version += f" {channel[v[3]]}"
            if v[4] != 0:
                version += f" {v[4]}"
        return self._version

    def get_archive_names(self) -> list[str]:
        proc_arch = ProcessorArchitecture.from_python()
        assert proc_arch == ProcessorArchitecture.from_os()

        lookup_os_name = platform.system()
        str_arch = proc_arch.get_machine_repr()
        qt_name = ""
        with suppress(Exception):
            import qtpy

            qt_name = qtpy.API_NAME
        if lookup_os_name == "Windows":
            return [
                f"{self.filestem}_Win_{str_arch}.zip",
                f"{self.filestem}_Windows_{str_arch}.zip",
                f"{self.filestem}_Windows_{qt_name}_{str_arch}.zip",
            ]
        if lookup_os_name == "Linux":
            return [
                f"{self.filestem}_Linux_{str_arch}.zip",
                f"{self.filestem}_Linux_{str_arch}.tar.gz",
                f"{self.filestem}_Linux_{str_arch}.tar.bz2",
                f"{self.filestem}_Linux_{str_arch}.tar.xz",
                f"{self.filestem}_Linux_{qt_name}_{str_arch}.zip",
                f"{self.filestem}_Linux_{qt_name}_{str_arch}.tar.gz",
                f"{self.filestem}_Linux_{qt_name}_{str_arch}.tar.bz2",
                f"{self.filestem}_Linux_{qt_name}_{str_arch}.tar.xz",
            ]
        if lookup_os_name == "Darwin":
            return [
                f"{self.filestem}_Mac_{str_arch}.tar.gz",
                f"{self.filestem}_macOS_{str_arch}.tar.gz",
                f"{self.filestem}_Mac_{str_arch}.tar.bz2",
                f"{self.filestem}_macOS_{str_arch}.tar.bz2",
                f"{self.filestem}_Mac_{str_arch}.tar.xz",
                f"{self.filestem}_macOS_{str_arch}.tar.xz",
                f"{self.filestem}_Mac_{str_arch}.zip",
                f"{self.filestem}_macOS_{str_arch}.zip",
                f"{self.filestem}_macOS_{qt_name}_{str_arch}.zip",
                f"{self.filestem}_macOS_{qt_name}_{str_arch}.tar.gz",
                f"{self.filestem}_macOS_{qt_name}_{str_arch}.tar.bz2",
                f"{self.filestem}_macOS_{qt_name}_{str_arch}.tar.xz",
            ]

        raise ValueError(f"Unexpected and unsupported OS: {lookup_os_name}")

    def is_downloaded(self) -> bool | None:
        """Used to check if update has been downloaded.

        ######Returns (bool):
            True - File is already downloaded.
            False - File has not been downloaded.
        """
        return False if self._is_downloading else self._is_downloaded()

    def download(
        self,
        *,
        background: bool = False,
    ) -> bool | None:
        """Downloads update.

        ######Args:
            background (bool): Perform download in background thread
        """
        if background:
            if not self._is_downloading:
                self._is_downloading = True
                threading.Thread(target=self._download, name="LibUpdate_download_thread").start()
        elif not self._is_downloading:
            self._is_downloading = True
            return self._download()
        return None

    def extract(self) -> bool:
        """Will extract the update from its archive to the update folder.
        If updating a lib you can take over from there. If updating
        an app this call should be followed by method "restart" to
        complete update.

        ######Returns:
            (bool)
                True - Extract successful.
                False - Extract failed.
        """
        try:
            self._extract_update()
        except OSError as err:
            self.log.debug(err, exc_info=True)
            return False
        return True

    def _download(self) -> bool:
        if self.filestem is not None:
            if self._is_downloaded():  # pragma: no cover
                self._download_status = True
            else:
                self.log.debug("Starting full download")
                update_success = self._full_update()
                if update_success:
                    self._download_status = True
                    self.log.debug("Full download successful")
                else:  # pragma: no cover
                    self.log.debug("Full download failed")

        self._is_downloading = False
        return self._download_status

    def _find_archive_path(self) -> Path | None:
        """Find the first existing archive path from the list of archive names.

        Returns
        -------
            Path | None: The path to the first existing archive, or None if none found.
        """
        with ChDir(self.update_folder):
            for archive_name in self.get_archive_names():
                archive_path = Path.cwd().joinpath(archive_name).absolute()
                if archive_path.is_file():
                    return archive_path
        return None

    def _extract_update(self):
        self.log.info("Main extraction, starting in working dir '%s'", self.update_folder)
        with ChDir(self.update_folder):
            archive_path = self._find_archive_path()
            if archive_path is not None:
                self.log.info("Found archive %s", archive_path.name)
                self._recursive_extract(archive_path)
            else:
                self.log.warning("Archive not found, attempting to find it via similar extensions")
                for archive_name in self.get_archive_names():
                    base_path = Path.cwd().joinpath(archive_name).absolute()
                    for ext in (".gz", ".tar", ".zip", ".bz2"):
                        test_path = base_path.with_suffix(ext)
                        if test_path.is_file():
                            self.log.info("Found archive %s", test_path.name)
                            self._recursive_extract(test_path)
                            return

    @classmethod
    def _recursive_extract(
        cls,
        archive_path: Path,
    ):
        log: Logger = RobustLogger()
        if not archive_path.is_file():
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(archive_path))
        if not os.access(str(archive_path), os.R_OK):
            raise PermissionError(errno.EACCES, os.strerror(errno.EACCES), str(archive_path))

        log.debug(f"(recursive) Extracting '{archive_path}'...")  # noqa: G004
        archive_ext = get_normalized_extension(archive_path)
        if archive_ext in {".gz", ".bz2", ".tar"}:
            cls.extract_tar(archive_path, recursive_extract=True)
        elif archive_ext == ".zip":
            cls.extract_zip(archive_path, recursive_extract=True)
        else:
            raise ValueError(f"Invalid file extension: '{archive_ext}' for archive path '{archive_path}'")

    @classmethod
    def extract_tar(
        cls,
        archive_path: os.PathLike | str,
        *,
        recursive_extract: bool = False,
    ):
        log = RobustLogger()
        log.info("Extracting TAR/GZ/BZIP archive at path '%s'", archive_path)
        try:
            with tarfile.open(archive_path, "r:*") as tfile:
                for member in tfile.getmembers():
                    if not member.isfile():
                        continue  # Ignore directories and non-file members
                    # Sanitize and extract each file
                    member_path: tuple[str, ...] = PurePath(member.name).parts
                    sanitized_path: Path = Path.cwd() / PurePath(*[p for p in member_path if p not in ("/", "..", "")])
                    ensure_directory_exists(sanitized_path.parent)  # Ensure directory exists
                    with sanitized_path.open("wb") as f:
                        extracted_member: IO[bytes] | None = tfile.extractfile(member)
                        if not extracted_member:
                            raise ValueError(f"Issue extracting member '{member}' ({member!r}): extractfile(member) returned None")  # noqa: TRY301
                        f.write(extracted_member.read())
                    if sanitized_path.exists():
                        log.info("Successfully extracted '%s'", sanitized_path)
                    else:
                        log.warning("Expected '%s' to exist after extraction, but it does not.", sanitized_path)
                    if not recursive_extract:
                        continue
                    if get_normalized_extension(sanitized_path) in {".gz", ".bz2", ".tar", ".zip"} and sanitized_path.is_file():
                        cls._recursive_extract(sanitized_path)
        except Exception as err:  # pragma: no cover
            log = RobustLogger()
            log.debug(err, exc_info=True)
            raise ValueError(f"Error reading tar/gzip file: {archive_path}") from err

    @classmethod
    def extract_zip(
        cls,
        archive_path: os.PathLike | str,
        *,
        recursive_extract: bool = False,
    ):
        log = RobustLogger()
        log.info("Extracting ZIP '%s'", archive_path)
        try:
            with zipfile.ZipFile(archive_path, "r") as zfile:
                zfile.extractall()  # noqa: S202
                if not recursive_extract:
                    return
                for file_info in zfile.infolist():
                    extracted_path = Path.cwd() / file_info.filename
                    if extracted_path.exists():
                        log.info("Successfully extracted '%s'", extracted_path)
                    else:
                        log.warning("Expected '%s' to exist after extraction, but it does not.", extracted_path)
                    if get_normalized_extension(extracted_path) in {".gz", ".bz2", ".tar", ".zip"} and extracted_path.is_file():
                        cls._recursive_extract(extracted_path)
        except Exception as err:  # pragma: no cover
            log.debug(err, exc_info=True)
            raise ValueError("Error reading zip file") from err

    def _is_downloaded(self) -> bool | None:
        """Checks if latest update is already downloaded."""
        # TODO(th3w1zard1): Compare file hashes to ensure security
        return self._find_archive_path() is not None

    def _full_update(self) -> bool:
        self.log.debug("Starting full update")
        result = True
        with ChDir(self.update_folder):
            self.log.debug("Downloading update...")
            archive_path: Path = Path(self.get_archive_names()[0]).absolute()
            parsed_url: str = ""
            for url in self.update_urls:
                parsed_url = url
                try:
                    if self.downloader:
                        return self.downloader(self.archive_name, parsed_url, self.progress_hooks)

                    if "mega.nz" in parsed_url.lower():
                        download_mega_file_url(parsed_url, archive_path, progress_hooks=self.progress_hooks)
                    else:
                        # HACK(th3w1zard1): use the latest tag for GitHub based downloads.
                        tag: str = self.latest
                        if self.version_to_tag_parser is not None:
                            tag = self.version_to_tag_parser(tag)
                        parsed_url = parsed_url.replace("{tag}", tag)
                        fd = FileDownloader(
                            self.archive_name,
                            [parsed_url],
                            hexdigest=None,
                            verify=True,
                            progress_hooks=self.progress_hooks,
                            max_download_retries=self.max_download_retries,
                            headers=None,
                            http_timeout=self.http_timeout,
                        )
                        result = fd.download_verify_write()
                        if result:
                            self.log.debug("Download Complete")
                            archive_path = fd.filepath.parent / fd.downloaded_filename  # TODO(th3w1zard1): loop through archive names...
                            if not archive_path.is_file():
                                self.log.error("Download complete but %s not found on disk at expected %s filepath?", fd.downloaded_filename, archive_path)
                                continue
                            if fd.downloaded_filename not in self.get_archive_names():
                                self.log.warning("Archive name %s does not exist in the archive name getter function, will override with downloaded name.", archive_path.name)

                                def getarchivename(a: Path = archive_path) -> list[str]:
                                    return [a.name]

                                self.get_archive_names = getarchivename
                        else:  # pragma: no cover
                            self.log.debug("Failed To Download Latest Version")
                    if archive_path.is_file():
                        break  # One of the mirrors worked successfully.
                except Exception:  # noqa: PERF203
                    self.log.exception("Exception while downloading %s", url)
        # if not archive_path.is_file():
        # exc = FileNotFoundError()
        # exc.filename = str(archive_path)
        # exc.strerror = "file downloader finished, but archive filepath doesn't exist."
        return bool(result and archive_path.is_file())

    def cleanup(self):
        shutil.rmtree(self.update_temp_path, ignore_errors=True)


class AppUpdate(LibUpdate):  # pragma: no cover
    """Used to update an application. This object is returned by `update_check`."""

    def __init__(
        self,
        update_urls: list[str],
        filestem: str,
        current_version: str,
        latest: str,
        progress_hooks: list[Callable[[dict[str, Any]], Any]] | None = None,
        max_download_retries: int | None = None,
        downloader: Callable | None = None,
        http_timeout: int | None = None,
        u_strategy: UpdateStrategy = UpdateStrategy.RENAME,
        r_strategy: RestartStrategy = RestartStrategy.DEFAULT,
        exithook: Callable | None = None,
        version_to_tag_parser: Callable | None = None,
    ):
        super().__init__(
            update_urls,
            filestem,
            current_version,
            latest,
            progress_hooks,
            max_download_retries,
            downloader,
            http_timeout,
            u_strategy,
            r_strategy,
            None,
            version_to_tag_parser,
        )
        self.exithook: Callable | None = exithook

    def extract_restart(self):
        """Will extract the update, overwrite the current binary, then restart the application using the updated binary."""
        try:
            self._extract_update()
            if os.name == "nt":
                if self.u_strategy == UpdateStrategy.RENAME:
                    self._win_rename(restart=True)
                else:
                    self._win_overwrite(restart=True)
            else:
                self._unix_overwrite()
                self._unix_restart()
        except OSError as err:
            self.log.debug(err, exc_info=True)

    def extract_overwrite(self):
        """Will extract the update then overwrite the current binary."""
        try:
            self._extract_update()
            if os.name == "nt":
                if self.u_strategy == UpdateStrategy.RENAME:
                    self._win_rename()
                else:
                    self._win_overwrite()
            else:
                self._unix_overwrite()
        except OSError as err:
            self.log.debug(err, exc_info=True)

    def _unix_overwrite(self):
        # Unix: Overwrites the running applications binary
        if platform.system() == "Darwin" and self._current_app_dir.name.endswith("MacOS"):
            self.log.debug("Looks like we're dealing with a Mac GUI")
            temp_dir: Path = get_mac_dot_app_dir(self._current_app_dir)
            self._current_app_dir = temp_dir

        app_update_path: Path = Path(self.update_folder, self.filename)
        self.log.debug("Update Location:\n%s", app_update_path.parent)
        self.log.debug("Update Name: %s", app_update_path.parent.name)

        current_app_path: Path = Path(self._current_app_dir, self.filename)
        self.log.debug("Current App location: %s", current_app_path)

        # Remove current app to prevent errors when moving update to new location
        # if update_app is a directory, then we are updating a directory
        if app_update_path.is_dir():
            if current_app_path.is_dir():
                shutil.rmtree(str(current_app_path))
            else:
                current_app_path.unlink()

        remove_any(current_app_path)

        self.log.info("Moving update: %s --> %s", app_update_path, self._current_app_dir)
        shutil.move(str(app_update_path), str(current_app_path))

    def _unix_restart(self):
        self.log.debug("Restarting %s", self.filename)
        app_path: Path = Path(self._current_app_dir, self.filename)
        if platform.system() == "Darwin" and get_normalized_extension(app_path) == ".app":
            self.log.debug(f"Must be a .app bundle: '{app_path}'")  # noqa: G004
            mac_app_binary_dir: Path = app_path.joinpath("Contents", "MacOS")

            # We are making an assumption here that only 1
            # executable will be in the MacOS folder.
            app_path = mac_app_binary_dir / self.filestem

        r = Restarter(
            app_path,
            app_path,
            restart_strategy=self.r_strategy,
            filename=self.filestem,
            update_strategy=self.u_strategy,
            exithook=self.exithook,
        )
        r.process()

    def _win_rename(
        self,
        *,
        restart: bool = False,
    ) -> tuple[Path, Path]:
        """This function renames the current application with a temporary name and moves the updated application into its place.

        It also handles rollback in case of failure.

        Args:
        ----
            - restart (bool): If True, the updated application will be launched after the rename process is complete. Default is False.

        Returns:
        -------
            - tuple[Path, Path]: A tuple containing the paths of the old application and the current application after the rename process is complete.

        Processing Logic:
        ----------------
            - Checks if the current application path is a directory and raises an error if it is.
            - Renames the current application with a temporary name.
            - Checks if the updated application is a file and raises an error if it is not.
            - Removes the old application from previous updates.
            - Renames the updated application to the current application name.
            - Hides the old application.
            - Launches the updated application if restart is True.
            - Handles rollback in case of failure.
        """
        exe_name = self.filename
        cur_app_filepath = self._current_app_dir / exe_name
        if cur_app_filepath.is_dir():
            raise ValueError(f"Current app path '{cur_app_filepath}' cannot be a directory!")

        old_app_path = self._current_app_dir / f"{exe_name}.old"
        temp_app_filepath = self.update_temp_path.joinpath(exe_name)

        # Ensure it's a file.
        if not temp_app_filepath.is_file():
            # Detect if archive expanded some folder with the same name as the archive.
            # This assumes all dots in the archive filename are extensions of the file archive.
            temp_app_filepath = temp_app_filepath.parent
            for archive_name in self.get_archive_names():
                archive_stem = archive_name[: archive_name.find(".")]
                check_path = temp_app_filepath.joinpath(archive_stem, exe_name)
                if check_path.is_file():
                    temp_app_filepath = check_path
                    break
            if temp_app_filepath.is_dir():
                self.log.warning("The rename strategy is only supported for one file bundled executables. Falling back to overwrite strategy.")
                self._win_overwrite(restart=True)
            elif not temp_app_filepath.exists():
                self.log.error("Updated app filepath: %s not found", temp_app_filepath)

        # Remove the old app from previous updates
        try:
            if old_app_path.is_file():
                old_app_path.unlink(missing_ok=True)
            elif old_app_path.is_dir():
                shutil.rmtree(str(old_app_path), ignore_errors=True)
            elif not old_app_path.exists():
                raise ValueError(f"Old app path at '{old_app_path}' was neither a file or directory, perhaps we don't have permission to check? No changes have been made.")
        except PermissionError:
            # Fallback to the good ol' rename strategy.
            randomized_old_app_path: Path = old_app_path.with_suffix(f"{old_app_path.suffix}.{uuid.uuid4().hex[:7]}")
            old_app_path.rename(randomized_old_app_path)

        # On Windows, it's possible to rename a currently running exe file
        if is_frozen() or cur_app_filepath.exists():  # exe may not exist if running from .py source
            cur_app_filepath.rename(old_app_path)

        # Any operation from here forward will require rollback on failure
        try:
            self.log.info("Copy '%s' --> '%s'", temp_app_filepath, cur_app_filepath)
            shutil.copy(str(temp_app_filepath), str(cur_app_filepath))
        except OSError:
            self.log.exception("Failed to move updated app into position, rolling back")
            self._win_rollback_on_exception(cur_app_filepath, old_app_path)

        assert temp_app_filepath.is_file()
        if is_frozen() or old_app_path.is_file():  # exe may not exist if running from .py source
            try:
                win_hide_file(old_app_path)
            except OSError:
                self.log.info("Failed to hide file. This is fine - we can still continue", exc_info=True)

        if not restart:
            return old_app_path, cur_app_filepath

        try:
            r = Restarter(
                cur_app_filepath, temp_app_filepath, restart_strategy=self.r_strategy, filename=self.filename, update_strategy=self.u_strategy, exithook=self.exithook
            )
            r.process()
        except OSError as e:
            if not is_frozen():
                raise  # nothing to roll back if working from src
            # Raised by os.execl
            self.log.exception("Failed to launch updated app, rolling back")
            # Rollback strategy: unhide old app, delete current app, move old app back
            try:
                import ctypes

                attrs: int = ctypes.windll.kernel32.GetFileAttributesW(str(old_app_path))
                if attrs == -1:
                    raise ctypes.WinError() from e
                if not ctypes.windll.kernel32.SetFileAttributesW(str(old_app_path), attrs & (~0x02)):
                    raise ctypes.WinError() from e
            except OSError:
                # Better to stay hidden than to just fail at this point
                self.log.exception("Could not unhide file in rollback process")
            self._win_rollback_on_exception(cur_app_filepath, old_app_path)
        return old_app_path, cur_app_filepath

    def _win_rollback_on_exception(
        self,
        cur_app_filepath: Path,
        old_app_path: Path,
        *,
        no_raise: bool = False,
    ):
        """Rollback strategy: move current app back into position."""
        cur_app_filepath.unlink(missing_ok=True)
        old_app_path.rename(cur_app_filepath)
        if no_raise:
            return
        # Check if there is an exception currently being handled
        # If there is an exception, re-raise it
        exc_type, exc_value, _exc_traceback = sys.exc_info()
        if exc_type is not None and exc_value is not None:
            raise exc_value

    def _win_overwrite(
        self,
        *,
        restart: bool = False,
    ):
        """Moves update to current directory of running application then restarts application using new update."""
        update_folder_path: Path = Path(self.update_folder)
        current_app_path: Path = self._current_app_dir / self.filename
        archive_stem: str = self.archive_name[: self.archive_name.find(".")]

        # TODO(th3w1zard1): clean this up... i'm ashamed.
        check_path1: Path = update_folder_path.joinpath(archive_stem)
        check_path2: Path = update_folder_path.joinpath(archive_stem.replace("_Win-", "_Windows_"))

        # Detect if archive expanded some folder with the same name as the archive.
        # This assumes all dots in the archive filename are extensions of the file archive.
        if check_path1.is_dir():
            updated_app_path: Path = check_path1
        elif check_path2.is_dir():
            updated_app_path = check_path2
        else:
            updated_app_path = Path(self.update_folder, self.get_expected_filename())

        r = Restarter(
            current_app_path,
            updated_app=updated_app_path,
            restart_strategy=self.r_strategy,
            filename=self.filestem,
            update_strategy=self.u_strategy,
            exithook=self.exithook,
        )
        r.process()



"""Archive filtering helpers for PyKotor CLI.

PyKotor CLI users commonly think in terms of filenames like ``p_cand.utc`` rather than
separate (resref, restype) fields. These helpers implement a small, consistent rule
set that works across KEY/BIF/RIM/ERF listings and extraction.
"""

from __future__ import annotations

import fnmatch


def _has_glob(pattern: str) -> bool:
    return any(ch in pattern for ch in ("*", "?", "["))


def matches_resource_name(
    resref: str,
    ext: str,
    pattern: str,
) -> bool:
    """Return True if a resource matches the user-supplied filter.

    Rules:
    - If pattern contains glob chars (*, ?, [), use fnmatch (case-insensitive).
      - If pattern contains a dot, match against ``resref.ext``.
      - Otherwise, match against ``resref``.
    - If pattern contains a dot and no glob chars, treat it as an exact filename match
      against ``resref.ext`` (case-insensitive).
    - Otherwise (no dot, no glob chars), treat it as a prefix match on ``resref``
      (case-insensitive).
    """
    resref_l = resref.lower()
    ext_l = ext.lower()
    pat_l = pattern.lower()

    full_name = f"{resref_l}.{ext_l}"

    if _has_glob(pat_l):
        if "." in pat_l:
            return fnmatch.fnmatch(full_name, pat_l)
        return fnmatch.fnmatch(resref_l, pat_l)

    if "." in pat_l:
        return full_name == pat_l

    return resref_l.startswith(pat_l)



"""XML read/write for LIP (lip sync) keyframes; uses defusedxml when available."""

from __future__ import annotations

from typing import TYPE_CHECKING

# Try to import defusedxml, fallback to ET if not available
from xml.etree import ElementTree as ET

try:  # sourcery skip: remove-redundant-exception, simplify-single-exception-tuple
    from defusedxml.ElementTree import fromstring
except (ImportError, ModuleNotFoundError):
    from xml.etree import ElementTree as ET

    fromstring = ET.fromstring

from pykotor.resource.formats.lip.lip_data import LIP, LIPShape
from pykotor.resource.type import ResourceReader, ResourceWriter, autoclose
from utility.misc import indent

if TYPE_CHECKING:
    from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES


class LIPXMLReader(ResourceReader):
    """Reads LIP files from XML format.

    XML is a human-readable format for easier editing of lip-sync animation data.

    References:
    ----------
        See lip_data module docstring for engine addresses (K1 + TSL TODO).

        Note: XML format structure may vary between tools
    """

    def __init__(
        self,
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = 0,
    ):
        super().__init__(source, offset, size)
        self._lip: LIP | None = None

    @autoclose
    def load(self, *, auto_close: bool = True) -> LIP:  # noqa: FBT001, FBT002, ARG002
        self._lip = LIP()

        data: str = self._reader.read_bytes(self._reader.size()).decode()
        xml_root: ET.Element = fromstring(data)  # noqa: S314

        if xml_root.tag != "lip":
            msg = "The XML file that was loaded was not a valid LIP."
            raise ValueError(msg)

        duration: str | None = xml_root.get("duration")
        if duration is None:
            msg = "Missing duration of the LIP."
            raise ValueError(msg)
        self._lip.length = float(duration)

        for subelement in xml_root:
            time: str | None = subelement.get("time")
            if time is None:
                msg = "Missing time for a keyframe."
                raise ValueError(msg)

            shape: str | None = subelement.get("shape")
            if shape is None:
                msg = "Missing shape for a keyframe."
                raise ValueError(msg)
            self._lip.add(float(time), LIPShape(int(shape)))

        return self._lip


class LIPXMLWriter(ResourceWriter):
    def __init__(
        self,
        lip: LIP,
        target: TARGET_TYPES,
    ):
        super().__init__(target)
        self._lip: LIP = lip
        self._xml_root: ET.Element = ET.Element("lip")

    @autoclose
    def write(self, *, auto_close: bool = True):  # noqa: FBT001, FBT002, ARG002  # pyright: ignore[reportUnusedParameters]
        self._xml_root.set("duration", str(self._lip.length))

        for keyframe in self._lip:
            ET.SubElement(
                self._xml_root,
                "keyframe",
                time=str(keyframe.time),
                shape=str(keyframe.shape.value),
            )

        indent(self._xml_root)
        self._writer.write_bytes(ET.tostring(self._xml_root))



"""Console/terminal helpers for PyKotor CLI.

These helpers exist primarily to keep CLI output robust on Windows terminals that
may default to non-UTF-8 encodings (e.g. cp1252). Avoid emitting unicode glyphs
unless the active stdout encoding can represent them.
"""

from __future__ import annotations

import sys

from functools import lru_cache


def _stream_can_encode(text: str) -> bool:
    stream = sys.stdout
    encoding = getattr(stream, "encoding", None) or sys.getdefaultencoding()
    try:
        text.encode(encoding)
    except Exception:
        return False
    return True


@lru_cache(maxsize=1)
def ok_fail_symbols() -> tuple[str, str]:
    """Return `(ok, fail)` symbols that are safe for the current stdout encoding."""
    if _stream_can_encode("✓✗"):
        return "✓", "✗"
    return "OK", "X"



"""This module handles classes relating to editing ERF files.

ERF (Encapsulated Resource File) files are self-contained archives used for modules, save games,
texture packs, and hak paks. Unlike BIF files which require a KEY file for filename lookups,
ERF files store both resource names (ResRefs) and data in the same file. They also support
localized strings for descriptions in multiple languages.

References:
----------
        Based on unified K1 (swkotor.exe) and TSL (swkotor2.exe) ERF structure.
        Addresses: (K1: swkotor.exe, TSL: swkotor2.exe — verify/fill TSL via REVA when available).

        - CExoEncapsulatedFile::CExoEncapsulatedFile — constructor for encapsulated file.
          K1: 0x0040ef90, TSL: TODO
        - CExoKeyTable::AddEncapsulatedContents — adds ERF/MOD/SAV contents to key table; tries NWM, MOD, SAV, ERF; opens "rb"; reads header/entries.
          K1: 0x0040f3c0, TSL: TODO
          Stack/header: Type [0x00], Version [0x04] ("V1.0"), EntryCount [0x10], KeyOffset [0x18]; Build Year/Day/DescStrRef allocated but unused.
        - "MOD V1.0" string (MOD file version identifier): K1: 0x0074539c, TSL: TODO
        ERF file format specification
        Binary Format:
        -------------
        Header (160 bytes):
        Offset | Size | Type   | Description
        -------|------|--------|-------------
        0x00   | 4    | char[] | File Type ("ERF ", "MOD ", "SAV ", "HAK ")
        0x04   | 4    | char[] | File Version ("V1.0")
        0x08   | 4    | uint32 | Language Count
        0x0C   | 4    | uint32 | Localized String Size (total bytes)
        0x10   | 4    | uint32 | Entry Count (number of resources)
        0x14   | 4    | uint32 | Offset to Localized String List
        0x18   | 4    | uint32 | Offset to Key List
        0x1C   | 4    | uint32 | Offset to Resource List
        0x20   | 4    | uint32 | Build Year (years since 1900)
        0x24   | 4    | uint32 | Build Day (days since Jan 1)
        0x28   | 4    | uint32 | Description StrRef (TLK reference)
        0x2C   | 116  | byte[] | Reserved (padding, usually zeros)
        Localized String Entry (variable length per language):
        - 4 bytes: Language ID (see Language enum)
        - 4 bytes: String Size (length in bytes)
        - N bytes: String Data (windows-1252 encoded text)
        Key Entry (24 bytes each):
        Offset | Size | Type   | Description
        -------|------|--------|-------------
        0x00   | 16   | char[] | ResRef (filename, null-padded, max 16 chars)
        0x10   | 4    | uint32 | Resource ID (index into resource list)
        0x14   | 2    | uint16 | Resource Type
        0x16   | 2    | uint16 | Unused (padding)
        Resource Entry (8 bytes each):
        Offset | Size | Type   | Description
        -------|------|--------|-------------
        0x00   | 4    | uint32 | Offset to Resource Data
        0x04   | 4    | uint32 | Resource Size
        Resource Data:
        Raw binary data for each resource at specified offsets


    Reference: Original BioWare engine binaries (ERF from swkotor.exe, swkotor2.exe). See module docstring for K1/TSL addresses.
"""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from pykotor.common.misc import ResRef
from pykotor.resource.bioware_archive import ArchiveResource, BiowareArchive
from pykotor.resource.type import ResourceType
from pykotor.tools.misc import is_erf_file, is_mod_file, is_sav_file

if TYPE_CHECKING:
    import os

    from pykotor.common.misc import ResRef


class ERFResource(ArchiveResource):
    """A single resource stored in an ERF/MOD/SAV file.

    Unlike BIF resources, ERF resources include their ResRef (filename) directly in the
    archive. Each resource is identified by a unique ResRef and ResourceType combination.

    References:
    ----------
        See module docstring for engine addresses (K1 + TSL TODO). CExoEncapsulatedFile::CExoEncapsulatedFile, CExoKeyTable::AddEncapsulatedContents, "MOD V1.0".
        https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Formats/KotorERF/ERFBinaryStructure.cs - Key and Resource entries
        https://github.com/th3w1zard1/KotOR_IO/tree/master/KotOR_IO/File Formats/ERF.cs - Key and Resource classes
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/interface/resource/IERFKeyEntry.ts - Key entry interface
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/interface/resource/IERFResource.ts - Resource entry interface


    Attributes:
    ----------
        All attributes inherited from ArchiveResource (resref, restype, data, size)
        ERF resources have no additional attributes beyond the base ArchiveResource
    """

    def __init__(self, resref: ResRef, restype: ResourceType, data: bytes):
        # https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Formats/KotorERF/ERFBinaryStructure.cs:119-120
        # https://github.com/th3w1zard1/KotOR_IO/tree/master/KotOR_IO/File Formats/ERF.cs:197-198
        # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/ERFObject.ts:103-107
        # ResRef stored in Key Entry (16 bytes, null-padded)
        # ResourceType stored in Key Entry (2 bytes, uint16)
        # Resource data referenced via Resource Entry (offset + size)
        super().__init__(resref=resref, restype=restype, data=data)


class ERFType(Enum):
    """The type of ERF file based on file header signature.

    ERF files can have different type signatures depending on their purpose:
    - ERF: Generic encapsulated resource file (texture packs, etc.)
    - MOD: Module file (game areas/levels)
    - SAV: Save game file
    - HAK: Hak pak file (custom content, unused in KotOR)

    References:
    ----------
        See module docstring for engine addresses (K1 + TSL TODO). CExoEncapsulatedFile::CExoEncapsulatedFile, CExoKeyTable::AddEncapsulatedContents, "MOD V1.0".
        https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Formats/KotorERF/ERFBinaryStructure.cs - FileType field
        https://github.com/th3w1zard1/KotOR_IO/tree/master/KotOR_IO/File Formats/ERF.cs - FileType reading
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/ERFObject.ts - File type default

    """

    ERF = "ERF "  # Generic ERF archive (texture packs, etc.)
    MOD = "MOD "  # Module file (game levels/areas)

    @classmethod
    def from_extension(cls, ext_or_filepath: os.PathLike | str) -> ERFType:
        if is_erf_file(ext_or_filepath):
            return cls.ERF
        if is_mod_file(ext_or_filepath):
            return cls.MOD
        if is_sav_file(ext_or_filepath):  # .SAV files still use the 'MOD ' signature in its first 4 bytes of the file header
            return cls.MOD
        msg = f"Invalid ERF extension in filepath '{ext_or_filepath}'."
        raise ValueError(msg)


class ERF(BiowareArchive):
    """Represents an ERF/MOD/SAV file.

    ERF (Encapsulated Resource File) is a self-contained archive format used for game modules,
    save games, and resource packs. Unlike BIF+KEY pairs, ERF files contain both resource names
    and data in a single file, making them ideal for distributable content like mods.

    References:
    ----------
        See module docstring for engine addresses (K1 + TSL TODO). CExoEncapsulatedFile::CExoEncapsulatedFile, CExoKeyTable::AddEncapsulatedContents, "MOD V1.0".
        https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Formats/KotorERF/ERFBinaryStructure.cs - FileRoot class
        https://github.com/th3w1zard1/KotOR_IO/tree/master/KotOR_IO/File Formats/ERF.cs - Complete ERF implementation
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/ERFObject.ts - ERFObject class


    Attributes:
    ----------
        erf_type: File type signature (ERF, MOD, SAV, HAK)
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/ERFBinaryStructure.cs:73 (FileType property)
            Reference: https://github.com/th3w1zard1/KotOR_IO/tree/master/ERF.cs:46 (FileType field)
            Reference: https://github.com/th3w1zard1/KotOR.js/tree/master/ERFObject.ts:45 (fileType default)
            Determines intended use of the archive
            ERF = texture packs, MOD = game modules, SAV = save games

        is_save: Flag indicating if this is a save game ERF
            Reference: https://github.com/th3w1zard1/KotOR_IO/tree/master/ERF.cs:15-16 (save game comment)
            Save games use MOD signature but have different structure
            Affects how certain fields are interpreted (e.g., build date)
            PyKotor-specific flag for save game handling

        build_year: Years since 1900 (e.g., 103 = 2003)
            Reference: https://github.com/th3w1zard1/Kotor.NET/blob/master/Kotor.NET/Formats/KotorERF/ERFBinaryStructure.cs:84 (BuildYear)

        build_day: Day of the year (1-366)
            Reference: https://github.com/th3w1zard1/Kotor.NET/blob/master/Kotor.NET/Formats/KotorERF/ERFBinaryStructure.cs:85 (BuildDay)

        description_strref: TLK String Reference for module description
            Reference: ERF File Format Specification (Offset 0x28)
            Note: Kotor.NET stops reading at 0x24 (BuildDay), skipping this field.
            Defaults: -1 for MOD/NWM, 0 for SAV

        localized_strings: Dictionary providing descriptions in multiple languages (LanguageID -> String)
            Reference: ERF File Format Specification (Offsets 0x08, 0x0C, 0x14)
            Note: reone (erfreader.cpp:28) and Kotor.NET (ERFBinaryStructure.cs:100) skip these fields.
            Used primarily in MOD files for module names/loading screens
    """

    BINARY_TYPE = ResourceType.ERF
    ARCHIVE_TYPE: type[ArchiveResource] = ERFResource
    COMPARABLE_FIELDS = ("erf_type", "is_save_erf", "build_year", "build_day", "description_strref", "localized_strings")
    COMPARABLE_SET_FIELDS = ("_resources",)

    def __init__(
        self,
        erf_type: ERFType = ERFType.ERF,
        *,
        is_save: bool = False,
        build_year: int = 0,
        build_day: int = 0,
        description_strref: int = -1,
        localized_strings: dict[int, str] | None = None,
    ):
        super().__init__()

        # https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Formats/KotorERF/ERFBinaryStructure.cs:73
        # https://github.com/th3w1zard1/KotOR_IO/tree/master/KotOR_IO/File Formats/ERF.cs:46
        # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/ERFObject.ts:45
        # File type signature (ERF, MOD, SAV, HAK)
        self.erf_type: ERFType = erf_type

        # PyKotor-specific flag for save game handling
        # Save games use MOD signature but have different behavior
        self.is_save: bool = is_save

        self.build_year: int = build_year
        self.build_day: int = build_day
        self.description_strref: int = description_strref
        self.localized_strings: dict[int, str] = localized_strings if localized_strings is not None else {}

    @property
    def is_save_erf(self) -> bool:
        """Alias for ComparableMixin compatibility."""
        return self.is_save

    @is_save_erf.setter
    def is_save_erf(self, value: bool) -> None:
        self.is_save = value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.erf_type!r}, is_save={self.is_save}, desc_strref={self.description_strref})"

    def __eq__(self, other: object):
        from pykotor.resource.formats.rim import RIM  # Prevent circular imports  # noqa: PLC0415

        if not isinstance(other, (ERF, RIM)):
            return NotImplemented  # type: ignore[no-any-return]
        return set(self._resources) == set(other._resources)

    def __hash__(self) -> int:
        return hash((self.erf_type, tuple(self._resources), self.is_save, self.description_strref))

    def get_resource_offset(self, resource: ArchiveResource) -> int:
        """Return the byte offset of a resource's data in serialized archive order."""
        from pykotor.resource.formats.erf.io_erf import ERFBinaryWriter

        entry_count = len(self._resources)
        data_start = ERFBinaryWriter.FILE_HEADER_SIZE + ERFBinaryWriter.KEY_ELEMENT_SIZE * entry_count

        offset = data_start
        for res in self._resources:
            if res == resource:
                return offset
            offset += len(res.data)
        raise ValueError("Resource is not present in ERF resource list")




"""2DA format detection and auto read/write dispatch (binary, CSV, JSON)."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.twoda.io_twoda import TwoDABinaryReader, TwoDABinaryWriter
from pykotor.resource.formats.twoda.io_twoda_csv import TwoDACSVReader, TwoDACSVWriter
from pykotor.resource.formats.twoda.io_twoda_json import TwoDAJSONReader, TwoDAJSONWriter
from pykotor.resource.type import ResourceType

if TYPE_CHECKING:
    from pykotor.resource.formats.twoda.twoda_data import TwoDA
    from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES


def detect_2da(
    source: SOURCE_TYPES,
    offset: int = 0,
) -> ResourceType:  # sourcery skip: assign-if-exp, reintroduce-else
    """Returns what format the TwoDA data is believed to be in.

    This function performs a basic check and does not guarantee accuracy of the result or integrity of the data.

    Args:
    ----
        source: Source of the TwoDA data.
        offset: Offset into the data.

    Raises:
    ------
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.

    Returns:
    -------
        The format of the TwoDA data.
    """

    def check(
        first4,
    ):
        if first4 == "2DA ":
            return ResourceType.TwoDA
        if "{" in first4:
            return ResourceType.TwoDA_JSON
        if "," in first4:
            return ResourceType.TwoDA_CSV
        # if "<" in first4:
        #    return ResourceType.TwoDA_XML
        return ResourceType.INVALID

    file_format: ResourceType
    try:
        with BinaryReader.from_auto(source, offset) as reader:
            file_format = check(reader.read_string(4))
    except (FileNotFoundError, PermissionError, IsADirectoryError):
        raise
    except OSError:
        file_format = ResourceType.INVALID

    return file_format


def read_2da(
    source: SOURCE_TYPES,
    offset: int = 0,
    size: int | None = None,
    file_format: ResourceType | None = None,
) -> TwoDA:
    """Returns an TwoDA instance from the source.

    The file format (TwoDA, TwoDA_CSV, TwoDA_JSON) is automatically determined before parsing the data.

    Args:
    ----
        source: The source of the data.
        offset: The byte offset of the file inside the data.
        size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.
        file_format: The file format to use (ResourceType.TwoDA, ResourceType.TwoDA_CSV, ResourceType.TwoDA_JSON). If not specified, it will be detected automatically.

    Raises:
    ------
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.
        ValueError: If the file was corrupted or the format could not be determined.

    Returns:
    -------
        An TwoDA instance.
    """
    if file_format is None:
        file_format = detect_2da(source, offset)

    if file_format == ResourceType.INVALID:
        msg = "Failed to determine the format of the 2DA file."
        raise ValueError(msg)

    if file_format == ResourceType.TwoDA:
        return TwoDABinaryReader(source, offset, size or 0).load()
    if file_format == ResourceType.TwoDA_CSV:
        return TwoDACSVReader(source, offset, size or 0).load()
    if file_format == ResourceType.TwoDA_JSON:
        return TwoDAJSONReader(source, offset, size or 0).load()
    msg = "detect_2da failed unexpectedly"
    raise ValueError(msg)


def write_2da(
    twoda: TwoDA,
    target: TARGET_TYPES,
    file_format: ResourceType = ResourceType.TwoDA,
):
    """Writes the TwoDA data to the target location with the specified format.

    Currently, the supported formats are: TwoDA, TwoDA_CSV and TwoDA_JSON.

    Args:
    ----
        twoda: The TwoDA file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
    ------
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be written to the specified destination.
        ValueError: If the specified format was unsupported.
    """
    if file_format == ResourceType.TwoDA:
        TwoDABinaryWriter(twoda, target).write()
    elif file_format == ResourceType.TwoDA_CSV:
        TwoDACSVWriter(twoda, target).write()
    elif file_format == ResourceType.TwoDA_JSON:
        TwoDAJSONWriter(twoda, target).write()
    else:
        msg = "Unsupported format specified; use TwoDA, TwoDA_CSV or TwoDA_JSON."
        raise ValueError(msg)


def bytes_2da(
    twoda: TwoDA,
    file_format: ResourceType = ResourceType.TwoDA,
) -> bytes:
    """Returns the TwoDA data in the specified format (TwoDA, TwoDA_CSV or TwoDA_JSON) as a bytes object.

    This is a convenience method that wraps the write_2da() method.

    Args:
    ----
        twoda: The target TwoDA object.
        file_format: The file format.

    Raises:
    ------
        ValueError: If the specified format was unsupported.

    Returns:
    -------
        The TwoDA data.
    """
    data = bytearray()
    write_2da(twoda, data, file_format)
    return data



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable


class Observable:
    def __init__(self):
        self.callbacks: list[Callable] = []

    def subscribe(self, callback: Callable):
        self.callbacks.append(callback)

    def fire(self, message: object):
        for callback in self.callbacks:
            callback(message)



"""Unpack command implementation.

This module handles unpacking of module/ERF/hak files into JSON source files.
"""

from __future__ import annotations

import glob
import hashlib

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from argparse import Namespace
    from logging import Logger

from pykotor.cli.cfg_parser import load_config
from pykotor.resource.formats.erf import read_erf
from pykotor.resource.formats.gff import read_gff, write_gff
from pykotor.resource.formats.rim import read_rim
from pykotor.resource.type import ResourceType


def compute_file_hash(file_path: Path) -> str:
    """Compute SHA1 hash of a file."""
    sha1 = hashlib.sha1()
    with file_path.open("rb") as f:
        while chunk := f.read(8192):
            sha1.update(chunk)
    return sha1.hexdigest()


def cmd_unpack(args: Namespace, logger: Logger) -> int:
    """Handle unpack command - unpack a file into source tree.

    Args:
    ----
        args: Parsed command line arguments
        logger: Logger instance

    Returns:
    -------
        Exit code (0 for success, non-zero for error)
    """
    # Load configuration
    config = load_config(logger)
    if config is None:
        return 1

    # Determine target
    target_name = args.target
    target = config.get_target(target_name)
    if target is None:
        if target_name:
            logger.error(f"Target not found: {target_name}")
        else:
            logger.error("No default target found")
        return 1

    target_name = target.get("name", "unnamed")
    logger.info(f"Unpacking target: {target_name}")

    # Determine file to unpack
    unpack_file = args.unpack_file or args.file
    if unpack_file:
        file_path = Path(unpack_file)
    else:
        # Try to find the target's file
        target_file = config.resolve_target_value(target, "file")
        if not target_file:
            logger.error("No file specified and target has no file defined")
            return 1

        # Check in current directory first
        file_path = Path(target_file)
        if not file_path.exists():
            # Try to find in KOTOR installation directory
            # This would require Installation detection - for now just error
            logger.error(f"File not found: {target_file}")
            return 1

    if not file_path.exists():
        logger.error(f"File not found: {file_path}")
        return 1

    logger.info(f"Unpacking from: {file_path}")

    # Determine file type and unpack
    try:
        suffix = file_path.suffix.lower()
        archive = None

        if suffix in (".mod", ".erf", ".sav"):
            archive = read_erf(file_path)
        elif suffix == ".rim":
            archive = read_rim(file_path)
        else:
            logger.error(f"Unsupported file type: {suffix}")
            return 1

        if archive is None:
            logger.error("Failed to read archive")
            return 1

        logger.info(f"Archive contains {len(list(archive))} resources")

        # Get target rules
        rules = config.get_target_rules(target)
        sources = config.get_target_sources(target)

        # Create cache directory for tracking
        cache_dir = config.root_dir / ".pykotorcli" / "cache" / target_name
        cache_dir.mkdir(parents=True, exist_ok=True)

        # Track unpacked files
        unpacked_files = []

        # Unpack each resource
        for resource in archive:
            resref = resource.resname
            restype = resource.restype
            filename = f"{resref}.{restype.extension}"

            # Determine destination based on rules
            destination = None

            # Check existing source tree first
            include_patterns = sources.get("include", [])
            for pattern in include_patterns:
                # Expand glob pattern
                pattern_path = config.root_dir / pattern.replace("**", "*")
                matches = list(glob.glob(str(pattern_path), recursive=True))
                for match in matches:
                    match_path = Path(match)
                    if match_path.name == filename:
                        destination = match_path
                        break
                if destination:
                    break

            # If not found in source tree, check rules
            if not destination:
                for pattern, path_str in rules.items():
                    import fnmatch

                    if fnmatch.fnmatch(filename, pattern):
                        dest_dir = config.root_dir / path_str
                        dest_dir.mkdir(parents=True, exist_ok=True)
                        destination = dest_dir / filename
                        break

            # If still not found, put in unknown directory
            if not destination:
                unknown_dir = config.root_dir / "unknown"
                unknown_dir.mkdir(parents=True, exist_ok=True)
                destination = unknown_dir / filename
                logger.warning(f"No rule matched {filename}, placing in unknown/")

            # Get resource data
            resource_data = bytes(resource.data)

            # Convert to JSON if GFF format
            if restype.is_gff():
                # Convert GFF to JSON using PyKotor's JSON writer
                try:
                    gff = read_gff(resource_data)

                    # Write as JSON
                    json_dest = destination.with_suffix(destination.suffix + ".json")
                    write_gff(gff, json_dest, file_format=ResourceType.GFF_JSON)

                    destination = json_dest
                    logger.debug(f"Converted {filename} to JSON: {destination.relative_to(config.root_dir)}")
                except Exception as e:
                    logger.warning(f"Failed to convert {filename} to JSON: {e}")
                    # Fall back to binary write
                    destination.write_bytes(resource_data)
                    logger.debug(f"Extracted {filename} (binary fallback): {destination.relative_to(config.root_dir)}")
            else:
                # Write binary file for non-GFF resources
                destination.write_bytes(resource_data)
                logger.debug(f"Extracted {filename}: {destination.relative_to(config.root_dir)}")

            unpacked_files.append(destination)

        logger.info(f"Successfully unpacked {len(unpacked_files)} files")

        # Handle removeDeleted option
        if hasattr(args, "removeDeleted") and args.removeDeleted:
            logger.info("Checking for deleted files...")
            # This would compare against previous unpack and remove files not in current archive
            # Implementation would require tracking previous unpacks

    except Exception as e:
        logger.exception(f"Failed to unpack: {e}")
        return 1

    return 0



from __future__ import annotations

import os

from qtpy.QtWidgets import QApplication


def running_under_pytest() -> bool:
    # pytest always sets this while executing tests.
    return "PYTEST_CURRENT_TEST" in os.environ


def process_events_if_safe() -> None:
    """Process Qt events only when it is safe.

    pytest-qt runs its own event processing loop and can deadlock if production
    code calls QApplication.processEvents() re-entrantly during those waits.

    This helper intentionally becomes a no-op under pytest.
    """

    if running_under_pytest():
        return

    if os.environ.get("PYKOTOR_QT_SKIP_PROCESS_EVENTS") == "1":
        return

    app = QApplication.instance()
    if app is None:
        return

    QApplication.processEvents()



"""KEY/BIF index types and KEYDataFile abstraction for archive resources keyed by BIF."""

from __future__ import annotations

import struct

from abc import abstractmethod
from dataclasses import dataclass
from typing import BinaryIO

from pykotor.resource.bioware_archive import ArchiveResource
from pykotor.resource.type import ResourceType


@dataclass
class BIFResource:
    name: str
    type: ResourceType
    bif_index: int
    res_index: int


class KEYDataFile(ArchiveResource):
    @abstractmethod
    def get_internal_resource_count(self) -> int:
        """Return the number of internal resources (including unmerged ones)."""

    @abstractmethod
    def merge_key(
        self,
        key: KEYFile,
        data_file_index: int,
    ) -> None:
        """Merge information from the KEY into the data file.

        Without this step, this data file archive does not contain any
        resource names at all.

        Args:
            key: A KEYFile with information about this data file.
            data_file_index: The index this data file has within the KEY file.
        """


class KEYFile:
    """Reads KEY files for resource indexing.

    References:
    ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries


    """

    def __init__(
        self,
        key: BinaryIO,
    ) -> None:
        self._bifs: list[str] = []
        self._resources: list[BIFResource] = []
        self.load(key)

    def get_bifs(self) -> list[str]:
        """Return a list of all managed bifs."""
        return self._bifs

    def get_resources(self) -> list[BIFResource]:
        """Return a list of all containing resources."""
        return self._resources

    def load(
        self,
        key: BinaryIO,
    ) -> None:
        self._read_header(key)
        bif_offset: int = struct.unpack("<I", key.read(4))[0]
        self._read_bif_list(key, bif_offset)

        resource_offset: int = struct.unpack("<I", key.read(4))[0]
        self._read_res_list(key, resource_offset)

    def _read_header(
        self,
        key: BinaryIO,
    ) -> None:
        file_type: bytes = key.read(4)
        file_version: bytes = key.read(4)

        if file_type != b"KEY ":
            raise ValueError("Not a KEY file")

        if file_version not in [b"V1  ", b"V1.1"]:
            raise ValueError(f"Unsupported KEY file version: {file_version}")

        self.bif_count: int = struct.unpack("<I", key.read(4))[0]
        self.resource_count: int = struct.unpack("<I", key.read(4))[0]

    def _read_bif_list(
        self,
        key: BinaryIO,
        offset: int,
    ) -> None:
        key.seek(offset)
        for _ in range(self.bif_count):
            size: int = struct.unpack("<I", key.read(4))[0]
            data_offset: int = struct.unpack("<I", key.read(4))[0]

            current_pos: int = key.tell()
            key.seek(data_offset)
            bif_name: str = key.read(size).decode("ascii").rstrip("\0")
            self._bifs.append(bif_name)
            key.seek(current_pos)

    def _read_res_list(
        self,
        key: BinaryIO,
        offset: int,
    ) -> None:
        key.seek(offset)
        for _ in range(self.resource_count):
            name: str = key.read(16).decode("ascii").rstrip("\0")
            res_type: int = struct.unpack("<H", key.read(2))[0]
            res_id: int = struct.unpack("<I", key.read(4))[0]

            # Decompose resource_id into bif_index (upper 12 bits) and res_index (lower 20 bits)
            bif_index: int = res_id >> 20
            res_index: int = res_id & 0xFFFFF

            resource: BIFResource = BIFResource(name, ResourceType(res_type), bif_index, res_index)
            self._resources.append(resource)



"""Type definitions for MDL/MDX files.

This module contains ONLY enums, flags, and constants used across the MDL/MDX format stack.
The canonical runtime MDL object model (classes with __eq__/__hash__/methods) lives in
`pykotor.resource.formats.mdl.mdl_data`.

For working with MDL models, import from:
    from pykotor.resource.formats.mdl import MDL, MDLNode, MDLMesh, MDLAnimation, ...

Architecture:
    - mdl_types.py: Enums, flags, constants (THIS FILE)
    - mdl_data.py: Runtime classes (MDL, MDLNode, MDLMesh, etc.) with full functionality
    - io_mdl.py: Binary MDL/MDX reader/writer
    - io_mdl_ascii.py: ASCII MDL reader/writer (MDLOps-compatible format)
    - mdl_auto.py: Format detection and dispatch

References:
    Based on swkotor.exe MDL/MDX structure:
    - LoadModel @ (K1: 0x00464200, TSL: 0x0047a570) - Loads MDL model via IODispatcher::ReadSync
    - LoadModel2 @ (K1: 0x0061b380, TSL: 0x00669ea0) - Alternative model loading function
      * Reads MDL/MDX file pair
      * Converts MaxTree to Model via MaxTree::AsModel
      * Checks modelsList for duplicates by name
    - ".mdl" extension @ (K1: 0x00740ca8, TSL: N/A - inline string literal) - MDL file extension
    - ".mdx" extension @ (K1: 0x00743944, TSL: N/A - inline string literal) - MDX file extension
    - "mdl" resource type @ (K1: 0x0074dd7c, TSL: N/A - inline string literal) - MDL resource identifier
    - "mdx" resource type @ (K1: 0x0074dc6c, TSL: N/A - inline string literal) - MDX resource identifier
    - "MDL" string @ (K1: 0x0075fb48, TSL: N/A - inline string literal) - MDL format identifier
    - "MDX" string @ (K1: 0x0075fb44, TSL: N/A - inline string literal) - MDX format identifier
    - Original BioWare engine binaries (swkotor.exe, swkotor2.exe)
    https://github.com/th3w1zard1/mdlops/tree/master/MDLOpsM.pm - Comprehensive MDL/MDX format constants and type definitions (tool)
    https://github.com/th3w1zard1/kotorblender/tree/master/io_scene_kotor/format/mdl/types.py - Blender MDL type definitions (tool)
     - KotOR MDL format specification (documentation)
"""

from __future__ import annotations

from enum import IntEnum, IntFlag


# region Enums and Flags
class MDLGeometryType(IntEnum):
    """Model geometry type indicating how it should be rendered."""

    GEOMETRY_UNKNOWN = 0
    GEOMETRY_NORMAL = 1
    GEOMETRY_SKINNED = 2
    GEOMETRY_DANGLY = 3
    GEOMETRY_SABER = 4

    # Back-compat aliases (older engine/tooling code used PascalCase names).
    # Keep these as enum aliases so `MDLGeometryType.Model` continues to work.
    Model = GEOMETRY_NORMAL
    Skinned = GEOMETRY_SKINNED
    Dangly = GEOMETRY_DANGLY
    Saber = GEOMETRY_SABER


class MDLClassification(IntEnum):
    """Model classification indicating its usage in the game."""

    INVALID = 0  # 0x0
    EFFECT = 1  # 0x1
    TILE = 2  # 0x2
    CHARACTER = 4  # 0x4
    DOOR = 8  # 0x8
    PLACEABLE = 16  # 0x10
    OTHER = 32  # 0x20
    GUI = 64  # 0x40
    ITEM = 128  # 0x80
    LIGHTSABER = 256  # 0x100
    WAYPOINT = 512  # 0x200
    WEAPON = 1024  # 0x400
    FURNITURE = 2048  # 0x800


class MDLNodeFlags(IntFlag):
    """Node flags indicating what type of data is attached to the node.

    These flags are combined to create specific node types. For example:
    - mesh = HEADER + MESH = 0x021 = 33
    - skin mesh = HEADER + MESH + SKIN = 0x061 = 97
    - dangly mesh = HEADER + MESH + DANGLY = 0x121 = 289
    - aabb mesh = HEADER + MESH + AABB = 0x221 = 545
    - saber mesh = HEADER + MESH + SABER = 0x821 = 2081

    References:
    - Based on swkotor.exe MDL node structure:
      * MdlNode::AsMdlNodeTriMesh @ (K1: 0x0043e400, TSL: 0x004501d0) - Casts node to tri-mesh (checks flags 0x21 = HEADER + MESH)
      * MdlNode::AsMdlNodeSkin @ (K1: 0x0043e3f0, TSL: 0x004501c0) - Casts node to skin mesh (checks flags 0x61 = HEADER + MESH + SKIN)
      * MdlNode::AsMdlNodeDanglyMesh @ (K1: 0x0043e380, TSL: 0x00450150) - Casts node to dangly mesh (checks flags 0x121 = HEADER + MESH + DANGLY)
      * MdlNode::AsMdlNodeLightsaber @ (K1: 0x0043e3a0, TSL: 0x00450170) - Casts node to lightsaber mesh (checks flags 0x821 = HEADER + MESH + SABER)
      * MdlNode::AsMdlNodeAABB @ (K1: 0x0043e340, TSL: 0x00450110) - Casts node to AABB mesh (checks flags 0x221 = HEADER + MESH + AABB)
      * MdlNode::AsMdlNodeEmitter @ (K1: 0x0043e3c0, TSL: 0x00450190) - Casts node to emitter (checks flags 5 = HEADER + EMITTER)
      * MdlNode::AsMdlNodeLight @ (K1: 0x0043e3d0, TSL: 0x004501a0) - Casts node to light (checks flags 3 = HEADER + LIGHT)
      * MdlNode::AsMdlNodeReference @ (K1: 0x0043e3e0, TSL: 0x004501b0) - Casts node to reference (checks flags 0x11 = HEADER + REFERENCE)
      * PartTriMesh::PartTriMesh @ (K1: 0x00445840, TSL: 0x00459be0) - Creates tri-mesh part from MDL node
      * LoadModel @ (K1: 0x00464200, TSL: 0x0047a570) - Loads MDL model via IODispatcher::ReadSync
      * LoadModel2 @ (K1: 0x0061b380, TSL: 0x00669ea0) - Alternative model loading function
    - Original BioWare engine binaries (swkotor.exe, swkotor2.exe)
    - https://github.com/th3w1zard1/mdlops/tree/master/MDLOpsM.pm:301-311 (Node type quick reference - tool)
    - https://github.com/th3w1zard1/kotorblender/tree/master/io_scene_kotor/format/mdl/types.py:93-101 (Node flags - tool)
    """

    # Base node flags
    # Reference: https://github.com/th3w1zard1/mdlops/tree/master/MDLOpsM.pm:302-311,
    HEADER = 0x0001  # NODE_HAS_HEADER - Base node data (mdlops:302)
    LIGHT = 0x0002  # NODE_HAS_LIGHT - Light data (mdlops:303, kotorblender:94)
    EMITTER = 0x0004  # NODE_HAS_EMITTER - Particle emitter data (mdlops:304, kotorblender:95)
    CAMERA = 0x0008  # NODE_HAS_CAMERA - Camera data (not in mdlops quick ref)
    REFERENCE = 0x0010  # NODE_HAS_REFERENCE - Reference to another model (mdlops:305, kotorblender:96)
    MESH = 0x0020  # NODE_HAS_MESH - Mesh geometry (mdlops:306, kotorblender:97)
    SKIN = 0x0040  # NODE_HAS_SKIN - Skinned mesh (mdlops:307, kotorblender:98)
    ANIM = 0x0080  # NODE_HAS_ANIM - Animation mesh (mdlops:308)
    DANGLY = 0x0100  # NODE_HAS_DANGLY - Cloth/hair physics mesh (mdlops:309, kotorblender:99)
    AABB = 0x0200  # NODE_HAS_AABB - Walkmes/collision data (mdlops:310, kotorblender:100)
    SABER = 0x0800  # NODE_HAS_SABER - Lightsaber blade mesh (mdlops:311, kotorblender:101)


class MDLNodeType(IntEnum):
    """Node types indicating the role of the node in the model."""

    DUMMY = 1  # Empty node for hierarchy
    TRIMESH = 2  # Basic mesh
    DANGLYMESH = 3  # Physics-enabled mesh
    LIGHT = 4  # Light source
    EMITTER = 5  # Particle emitter
    REFERENCE = 6  # Reference to another model
    PATCH = 7  # NURBS patch (unused)
    AABB = 8  # Axis-aligned bounding box
    SKIN = 9  # Skinned mesh (mdlops:320, NODE_SKIN = 97 = HEADER + MESH + SKIN flags)
    CAMERA = 10  # Camera viewpoint
    BINARY = 11  # Binary (unknown usage)
    SABER = 12  # Lightsaber blade


# region Controller Data
class MDLControllerType(IntEnum):
    """Controller types for animations and node properties.

    These controller types are used to animate various properties of nodes in MDL models.
    Controllers can be indexed by node type since some IDs are reused for different node types.

    References:
    - Based on swkotor.exe MDL controller structure:
      * Controller types are used to animate node properties (position, orientation, scale, alpha)
      * "scale" string @ (K1: 0x00741f44, TSL: N/A - inline string literal) - Scale controller identifier
      * "scalekey" string @ (K1: 0x00741f38, TSL: N/A - inline string literal) - Scale keyframe identifier
      * "scalebezierkey" string @ (K1: 0x00741f28, TSL: N/A - inline string literal) - Scale bezier keyframe identifier
      * "ALPHA" string @ (K1: 0x0073dfc0, TSL: N/A - inline string literal) - Alpha controller identifier
      * "channelscale" string @ (K1: 0x00741d1c, TSL: N/A - inline string literal) - Channel scale identifier
      * LoadModel @ (K1: 0x00464200, TSL: 0x0047a570) - Loads MDL models with controllers
      * LoadModel2 @ (K1: 0x0061b380, TSL: 0x00669ea0) - Alternative model loading function
    - Original BioWare engine binaries (swkotor.exe, swkotor2.exe)
    - https://github.com/th3w1zard1/mdlops/tree/master/MDLOpsM.pm:325-405 (Comprehensive controller mapping - tool)
    - https://github.com/th3w1zard1/kotorblender/tree/master/io_scene_kotor/format/mdl/types.py:140-197 (Controller constants - tool)

    Note: Controller indexing by node type is necessary because at least one controller ID (100)
    is used for different purposes in different node types (per https://github.com/th3w1zard1/mdlops/tree/master/MDLOpsM.pm:325)
    """

    INVALID = -1

    # Base node controllers (NODE_HAS_HEADER)
    # Reference: https://github.com/th3w1zard1/mdlops/tree/master/MDLOpsM.pm:329-335
    POSITION = 8  # position - All nodes (mdlops:329)
    ORIENTATION = 20  # orientation - All nodes (mdlops:330)
    SCALE = 36  # scale - All nodes (mdlops:331)
    ALPHA = 132  # alpha - All nodes (mdlops:332, was 128)

    # Light controllers (NODE_HAS_LIGHT)
    # Reference: https://github.com/th3w1zard1/mdlops/tree/master/MDLOpsM.pm:342-346,
    COLOR = 76  # color - Light nodes (mdlops:342)
    RADIUS = 88  # radius - Light nodes (mdlops:343)
    SHADOWRADIUS = 96  # shadowradius - Light nodes (mdlops:344)
    VERTICALDISPLACEMENT = 100  # verticaldisplacement - Light nodes (mdlops:345)
    MULTIPLIER = 140  # multiplier - Light nodes (mdlops:346)

    # Emitter controllers (NODE_HAS_EMITTER)
    # Reference: https://github.com/th3w1zard1/mdlops/tree/master/MDLOpsM.pm:357-405,
    # These mappings were updated based on fx_flame01.mdl analysis (mdlops:352-355)
    ALPHAEND = 80  # alphaEnd - Emitter (mdlops:357)
    ALPHASTART = 84  # alphaStart - Emitter (mdlops:358)
    BIRTHRATE = 88  # birthrate - Emitter (mdlops:359, same as RADIUS for lights)
    BOUNCECO = 92  # bounce_co - Emitter (mdlops:360)
    COMBINETIME = 96  # combinetime - Emitter, was 120 (mdlops:361)
    DRAG = 100  # drag - Emitter (mdlops:362, same as VERTICALDISPLACEMENT for lights)
    FPS = 104  # fps - Emitter, was 128 (mdlops:363)
    FRAMEEND = 108  # frameEnd - Emitter, was 132 (mdlops:364)
    FRAMESTART = 112  # frameStart - Emitter, was 136 (mdlops:365)
    GRAV = 116  # grav - Emitter, was 140 (mdlops:366)
    LIFEEXP = 120  # lifeExp - Emitter, was 144 (mdlops:367)
    MASS = 124  # mass - Emitter, was 148 (mdlops:368)
    P2P_BEZIER2 = 128  # p2p_bezier2 - Emitter, was 152 (mdlops:369)
    P2P_BEZIER3 = 132  # p2p_bezier3 - Emitter, was 156 (mdlops:370)
    PARTICLEROT = 136  # particleRot - Emitter, was 160 (mdlops:371)
    RANDVEL = 140  # randvel - Emitter, was 164 (mdlops:372)
    SIZESTART = 144  # sizeStart - Emitter, was 168 (mdlops:373)
    SIZEEND = 148  # sizeEnd - Emitter, was 172 (mdlops:374)
    SIZESTART_Y = 152  # sizeStart_y - Emitter, was 176 (mdlops:375)
    SIZEEND_Y = 156  # sizeEnd_y - Emitter, was 180 (mdlops:376)
    SPREAD = 160  # spread - Emitter, was 184 (mdlops:377)
    THRESHOLD = 164  # threshold - Emitter, was 188 (mdlops:378)
    VELOCITY = 168  # velocity - Emitter, was 192 (mdlops:379)
    XSIZE = 172  # xsize - Emitter, was 196 (mdlops:380)
    YSIZE = 176  # ysize - Emitter, was 200 (mdlops:381)
    BLURLENGTH = 180  # blurlength - Emitter, was 204 (mdlops:382)
    LIGHTNINGDELAY = 184  # lightningDelay - Emitter, was 208 (mdlops:383)
    LIGHTNINGRADIUS = 188  # lightningRadius - Emitter, was 212 (mdlops:384)
    LIGHTNINGSCALE = 192  # lightningScale - Emitter, was 216 (mdlops:385)
    LIGHTNINGSUBDIV = 196  # lightningSubDiv - Emitter (mdlops:386)
    LIGHTNINGZIGZAG = 200  # lightningzigzag - Emitter (mdlops:387)
    ALPHAMID = 216  # alphaMid - Emitter, was 464 (mdlops:388)
    PERCENTSTART = 220  # percentStart - Emitter, was 480 (mdlops:389)
    PERCENTMID = 224  # percentMid - Emitter, was 481 (mdlops:390)
    PERCENTEND = 228  # percentEnd - Emitter, was 482 (mdlops:391)
    SIZEMID = 232  # sizeMid - Emitter, was 484 (mdlops:392)
    SIZEMID_Y = 236  # sizeMid_y - Emitter, was 488 (mdlops:393)
    RANDOMBIRTHRATE = 240  # m_fRandomBirthRate - Emitter (mdlops:394)
    TARGETSIZE = 252  # targetsize - Emitter (mdlops:395)
    NUMCONTROLPTS = 256  # numcontrolpts - Emitter (mdlops:396)
    CONTROLPTRADIUS = 260  # controlptradius - Emitter (mdlops:397)
    CONTROLPTDELAY = 264  # controlptdelay - Emitter (mdlops:398)
    TANGENTSPREAD = 268  # tangentspread - Emitter (mdlops:399)
    TANGENTLENGTH = 272  # tangentlength - Emitter (mdlops:400)
    COLORMID = 284  # colorMid - Emitter, was 468 (mdlops:401)
    COLOREND = 380  # colorEnd - Emitter, was 96 (mdlops:402)
    COLORSTART = 392  # colorStart - Emitter, was 108 (mdlops:403)
    DETONATE = 502  # detonate - Emitter, was 228 (mdlops:404)

    # Mesh controllers (NODE_HAS_MESH)
    # Reference: https://github.com/th3w1zard1/mdlops/tree/master/MDLOpsM.pm:406,
    SELFILLUMCOLOR = 100  # selfillumcolor - Mesh (mdlops:406, same as DRAG and VERTICALDISPLACEMENT)

    # Legacy aliases for backward compatibility
    BOUNCE_CO = BOUNCECO
    P2P_BEZIER_2 = P2P_BEZIER2
    ILLUM_COLOR = SELFILLUMCOLOR
    ILLUM = SELFILLUMCOLOR
    SIZESTART_emitter = SIZESTART
    SIZEEND_emitter = SIZEEND
    LIGHTNINGDELAY_emitter = LIGHTNINGDELAY
    LIGHTNINGRADIUS_emitter = LIGHTNINGRADIUS
    LIGHTNINGSCALE_emitter = LIGHTNINGSCALE
    ALPHAMID_emitter = ALPHAMID
    SIZEMID_emitter = SIZEMID
    DETONATE_emitter = DETONATE

    # Additional back-compat aliases (PascalCase) used by PyKotorEngine tests/tooling.
    Position = POSITION
    Orientation = ORIENTATION
    Scale = SCALE
    Alpha = ALPHA
    Color = COLOR
    Radius = RADIUS
    Multiplier = MULTIPLIER

    # NOTE: Some controller IDs that were previously defined have been removed or
    # corrected based on vendor analysis. See fx_flame01.mdl analysis in mdlops:352-355


class MDLTrimeshProps(IntFlag):
    """Properties for trimesh nodes."""

    NONE = 0x00
    LIGHTMAP = 0x01
    COMPRESSED = 0x02
    UNKNOWN = 0x04
    TANGENTS = 0x08
    UNKNOWNA = 0x10
    UNKNOWNB = 0x20
    UNKNOWNC = 0x40
    UNKNOWND = 0x80


class MDLEmitterType(IntEnum):
    """Types of particle emitters."""

    STATIC = 0
    FIRE = 1
    FOUNTAIN = 2
    LIGHTNING = 3


class MDLRenderType(IntEnum):
    """Particle rendering types."""

    NORMAL = 0
    LINKED = 1
    BILLBOARD_TO_LOCAL_Z = 2
    BILLBOARD_TO_WORLD_Z = 3
    ALIGNED_TO_WORLD_Z = 4
    ALIGNED_TO_PARTICLE_DIR = 5
    MOTION_BLUR = 6


class MDLBlendType(IntEnum):
    """Particle blend modes."""

    NORMAL = 0
    PUNCH = 1
    LIGHTEN = 2
    MULTIPLY = 3


class MDLUpdateType(IntEnum):
    """Particle update modes."""

    FOUNTAIN = 0
    SINGLE = 1
    EXPLOSION = 2
    LIGHTNING = 3


class MDLTrimeshFlags(IntFlag):
    """Additional trimesh flags from KotOR implementation.

    References:
    - Based on swkotor.exe MDL trimesh structure:
      * PartTriMesh::PartTriMesh @ (K1: 0x00445840, TSL: 0x00459be0) - Creates tri-mesh part from MDL node
      * MdlNode::AsMdlNodeTriMesh @ (K1: 0x0043e400, TSL: 0x004501d0) - Casts node to tri-mesh (checks flags 0x21)
      * LoadModel @ (K1: 0x00464200, TSL: 0x0047a570) - Loads MDL models with trimesh flags
      * LoadModel2 @ (K1: 0x0061b380, TSL: 0x00669ea0) - Alternative model loading function
    - Original BioWare engine binaries (swkotor.exe, swkotor2.exe)
    """

    TILEFADE = 0x0001  # Has tile fade
    HEAD = 0x0002  # Is a head mesh
    RENDER = 0x0004  # Should be rendered
    SHADOW = 0x0008  # Casts shadows
    BEAMING = 0x0010  # Has beaming
    RENDER_ENV_MAP = 0x0020  # Should render environment mapping
    LIGHTMAP = 0x0040  # Has lightmap
    SKIN = 0x0080  # Is skinned mesh


class MDLLightFlags(IntFlag):
    """Light flags from KotOR implementation.

    References:
    - Based on swkotor.exe MDL light structure:
      * MdlNode::AsMdlNodeLight @ (K1: 0x0043e3d0, TSL: 0x004501a0) - Casts node to light
      * LoadModel @ (K1: 0x00464200, TSL: 0x0047a570) - Loads MDL models with light flags
      * LoadModel2 @ (K1: 0x0061b380, TSL: 0x00669ea0) - Alternative model loading function
    - Original BioWare engine binaries (swkotor.exe, swkotor2.exe)
    """

    ENABLED = 0x0001  # Light is enabled
    SHADOW = 0x0002  # Light casts shadows
    FLARE = 0x0004  # Light has lens flare
    FADING = 0x0008  # Light is fading
    AMBIENT = 0x0010  # Light is ambient only


class MDLEmitterFlags(IntFlag):
    """Particle emitter behavior flags.

    These flags control various aspects of particle emitter behavior including physics,
    inheritance, and rendering properties.

    References:
    - Based on swkotor.exe MDL emitter structure:
      * MdlNode::AsMdlNodeEmitter @ (K1: 0x0043e3c0, TSL: 0x00450190) - Casts node to emitter
      * LoadModel @ (K1: 0x00464200, TSL: 0x0047a570) - Loads MDL models with emitter flags
      * LoadModel2 @ (K1: 0x0061b380, TSL: 0x00669ea0) - Alternative model loading function
    - Original BioWare engine binaries (swkotor.exe, swkotor2.exe)
    - https://github.com/th3w1zard1/kotorblender/tree/master/io_scene_kotor/format/mdl/types.py:115-127 (Comprehensive list - tool)
    - https://github.com/th3w1zard1/kotorblender/tree/master/io_scene_kotor/format/mdl/reader.py:295-306 (Flag parsing - tool)
    """

    #
    P2P = 0x0001  # EMITTER_FLAG_P2P - Point-to-point emitter (kotorblender:115, reone:36)
    P2P_SEL = 0x0002  # EMITTER_FLAG_P2P_SEL - P2P selection (kotorblender:116)
    P2P_BEZIER = 0x0002  # P2P with Bezier curves (reone:37, same as P2P_SEL)
    AFFECTED_WIND = 0x0004  # EMITTER_FLAG_AFFECTED_WIND - Affected by wind (kotorblender:117, reone:38)
    TINTED = 0x0008  # EMITTER_FLAG_TINTED - Particles are tinted (kotorblender:118, reone:39)
    BOUNCE = 0x0010  # EMITTER_FLAG_BOUNCE - Particles bounce on collision (kotorblender:119, reone:40)
    RANDOM = 0x0020  # EMITTER_FLAG_RANDOM - Random rotation (kotorblender:120, reone:41)
    INHERIT = 0x0040  # EMITTER_FLAG_INHERIT - Inherit parent orientation (kotorblender:121, reone:42)
    INHERIT_VEL = 0x0080  # EMITTER_FLAG_INHERIT_VEL - Inherit parent velocity (kotorblender:122, reone:43)
    INHERIT_LOCAL = 0x0100  # EMITTER_FLAG_INHERIT_LOCAL - Inherit local transform (kotorblender:123, reone:44)
    SPLAT = 0x0200  # EMITTER_FLAG_SPLAT - Splat on collision (kotorblender:124, reone:45)
    INHERIT_PART = 0x0400  # EMITTER_FLAG_INHERIT_PART - Inherit particle properties (kotorblender:125, reone:46)
    DEPTH_TEXTURE = 0x0800  # EMITTER_FLAG_DEPTH_TEXTURE - Use depth texture (kotorblender:126, reone:47)
    FLAG_13 = 0x1000  # Unknown flag 13 (kotorblender:127, reone:48)

    # Legacy aliases
    LOOP = P2P  # Backward compatibility


class MDLSaberFlags(IntFlag):
    """Lightsaber flags from xoreos KotOR implementation."""

    FLARE = 0x0001  # Has flare effect
    DYNAMIC = 0x0002  # Dynamic lighting
    TRAIL = 0x0004  # Has motion trail


class MDLDynamicType(IntEnum):
    """Dynamic types for lights."""

    STATIC = 0
    DYNAMIC = 1
    ANIMATED = 2


# endregion



"""TXI (texture info) binary reader/writer and coordinate modes."""

from __future__ import annotations

from enum import IntEnum
from typing import TYPE_CHECKING

from loggerplus import RobustLogger
from pykotor.resource.type import ResourceReader, ResourceWriter, autoclose

if TYPE_CHECKING:
    from pykotor.resource.formats.txi.txi_data import TXI, TXICommand
    from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES


class TXIReaderMode(IntEnum):
    """This class is used to store the modes of a texture."""

    NORMAL = 0
    UPPER_LEFT_COORDS = 1
    LOWER_RIGHT_COORDS = 2


class TXIBinaryReader(ResourceReader):
    """Reads TXI (Texture Information) files.

    TXI files contain texture metadata including blending modes, bump maps, animations,
    and other rendering properties for TPC textures.

    References:
    ----------
        See txi_data module docstring for engine addresses (K1 + TSL TODO). GetTXIInternal, ReleaseTXIInternal, CAuroraTXI, CResTXI, SetTxiData, GetTxiData, ".txi", "txi", "TXI".

    """

    def __init__(self, source: SOURCE_TYPES, offset: int = 0, size: int = 0):
        super().__init__(source, offset, size)
        from pykotor.resource.formats.txi.txi_data import TXI

        self._txi: TXI = TXI()

    @autoclose
    def load(self, *func_args, auto_close: bool = True, **func_kwargs) -> TXI:  # noqa: FBT001, FBT002, ARG002
        from pykotor.resource.formats.txi.txi_data import TXI, TXICommand, TXIFeatures

        self._txi.features = TXIFeatures()
        self._empty: bool = True
        mode: TXIReaderMode = TXIReaderMode.NORMAL
        cur_coords: int = 0
        txi_bytes: bytes = self._reader.read_all()

        for line in txi_bytes.decode("ascii", errors="ignore").splitlines():
            try:
                parsed_line: str = line.strip()
                if not parsed_line:
                    continue

                # Check if this line is a command (starts with a known TXI command)
                # This allows commands to interrupt coordinate parsing
                args: str
                raw_cmd, args = (
                    parsed_line.split(" ", maxsplit=1)
                    if " " in parsed_line
                    else (
                        parsed_line,
                        "",
                    )
                )
                parsed_cmd_str: str = raw_cmd.strip().upper()
                is_command: bool = parsed_cmd_str in TXICommand.__members__

                if mode == TXIReaderMode.UPPER_LEFT_COORDS:
                    if is_command:
                        # Command encountered, exit coordinate mode and process the command
                        mode = TXIReaderMode.NORMAL
                    else:
                        # Try to parse as coordinates
                        try:
                            parts = parsed_line.split()
                            coords: tuple[float, float, int] = (
                                float(parts[0].strip()),
                                float(parts[1].strip()),
                                int(parts[2].strip()),
                            )
                            if self._txi.features.upperleftcoords is not None and cur_coords < len(self._txi.features.upperleftcoords):
                                self._txi.features.upperleftcoords[cur_coords] = coords
                            cur_coords += 1
                            if self._txi.features.upperleftcoords is not None and cur_coords >= len(self._txi.features.upperleftcoords):
                                mode = TXIReaderMode.NORMAL
                            continue
                        except (ValueError, IndexError):
                            # Not a valid coordinate line, might be a command, exit coordinate mode
                            mode = TXIReaderMode.NORMAL
                            # Fall through to process as command or skip

                if mode == TXIReaderMode.LOWER_RIGHT_COORDS:
                    if is_command:
                        # Command encountered, exit coordinate mode and process the command
                        mode = TXIReaderMode.NORMAL
                    else:
                        # Try to parse as coordinates
                        try:
                            parts = parsed_line.split()
                            coords = (
                                float(parts[0].strip()),
                                float(parts[1].strip()),
                                int(parts[2].strip()),
                            )
                            if self._txi.features.lowerrightcoords is not None and cur_coords < len(self._txi.features.lowerrightcoords):
                                self._txi.features.lowerrightcoords[cur_coords] = coords
                            cur_coords += 1
                            if self._txi.features.lowerrightcoords is not None and cur_coords >= len(self._txi.features.lowerrightcoords):
                                mode = TXIReaderMode.NORMAL
                            continue
                        except (ValueError, IndexError):
                            # Not a valid coordinate line, might be a command, exit coordinate mode
                            mode = TXIReaderMode.NORMAL
                            # Fall through to process as command or skip

                if not is_command:
                    RobustLogger().debug(f"Invalid TXI command: '{raw_cmd}'")
                    continue
                command: TXICommand = TXICommand.__members__[parsed_cmd_str]
                args = args.strip() if args else ""

                if command == TXICommand.ALPHAMEAN:
                    self._txi.features.alphamean = float(args)
                    self._empty = False
                elif command == TXICommand.ARTUROHEIGHT:
                    self._txi.features.arturoheight = int(args)
                    self._empty = False
                elif command == TXICommand.ARTUROWIDTH:
                    self._txi.features.arturowidth = int(args)
                    self._empty = False
                elif command == TXICommand.BASELINEHEIGHT:
                    self._txi.features.baselineheight = float(args)
                    self._empty = False
                elif command == TXICommand.BLENDING:
                    self._txi.features.blending = TXI.parse_blending(args)
                    self._empty = False
                elif command == TXICommand.BUMPMAPSCALING:
                    self._txi.features.bumpmapscaling = float(args)
                    self._empty = False
                elif command == TXICommand.BUMPMAPTEXTURE:
                    self._txi.features.bumpmaptexture = args
                    self._empty = False
                elif command == TXICommand.BUMPYSHINYTEXTURE:
                    self._txi.features.bumpyshinytexture = args
                    self._empty = False
                elif command == TXICommand.CANDOWNSAMPLE:
                    self._txi.features.candownsample = bool(int(args))
                    self._empty = False
                elif command == TXICommand.CARETINDENT:
                    self._txi.features.caretindent = float(args)
                    self._empty = False
                elif command == TXICommand.CHANNELSCALE:
                    self._txi.features.channelscale = list(map(float, args.split()))
                    self._empty = False
                elif command == TXICommand.CHANNELTRANSLATE:
                    self._txi.features.channeltranslate = list(map(float, args.split()))
                    self._empty = False
                elif command == TXICommand.CLAMP:
                    self._txi.features.clamp = bool(int(args))
                    self._empty = False
                elif command == TXICommand.CODEPAGE:
                    self._txi.features.codepage = int(args)
                    self._empty = False
                elif command == TXICommand.COLS:
                    self._txi.features.cols = int(args)
                    self._empty = False
                elif command == TXICommand.COMPRESSTEXTURE:
                    self._txi.features.compresstexture = bool(int(args))
                    self._empty = False
                elif command == TXICommand.CONTROLLERSCRIPT:
                    self._txi.features.controllerscript = args
                    self._empty = False
                elif command == TXICommand.CUBE:
                    self._txi.features.cube = bool(int(args))
                    self._empty = False
                elif command == TXICommand.DECAL:
                    self._txi.features.decal = bool(int(args))
                    self._empty = False
                elif command == TXICommand.DEFAULTBPP:
                    self._txi.features.defaultbpp = int(args)
                    self._empty = False
                elif command == TXICommand.DEFAULTHEIGHT:
                    self._txi.features.defaultheight = int(args)
                    self._empty = False
                elif command == TXICommand.DEFAULTWIDTH:
                    self._txi.features.defaultwidth = int(args)
                    self._empty = False
                elif command == TXICommand.DISTORT:
                    self._txi.features.distort = bool(int(args))
                    self._empty = False
                elif command == TXICommand.DISTORTANGLE:
                    self._txi.features.distortangle = float(args)
                    self._empty = False
                elif command == TXICommand.DISTORTIONAMPLITUDE:
                    self._txi.features.distortionamplitude = float(args)
                    self._empty = False
                elif command == TXICommand.DOWNSAMPLEFACTOR:
                    self._txi.features.downsamplefactor = float(args)
                    self._empty = False
                elif command == TXICommand.DOWNSAMPLEMAX:
                    self._txi.features.downsamplemax = int(args)
                    self._empty = False
                elif command == TXICommand.DOWNSAMPLEMIN:
                    self._txi.features.downsamplemin = int(args)
                    self._empty = False
                elif command == TXICommand.ENVMAPTEXTURE:
                    self._txi.features.envmaptexture = args
                    self._empty = False
                elif command == TXICommand.FILERANGE:
                    self._txi.features.filerange = list(map(int, args.split()))
                    self._empty = False
                elif command == TXICommand.FILTER:
                    self._txi.features.filter = bool(int(args))
                    self._empty = False
                elif command == TXICommand.FONTHEIGHT:
                    self._txi.features.fontheight = float(args)
                    self._empty = False
                elif command == TXICommand.FONTWIDTH:
                    self._txi.features.fontwidth = float(args)
                    self._empty = False
                elif command == TXICommand.FPS:
                    self._txi.features.fps = float(args)
                    self._empty = False
                elif command == TXICommand.ISBUMPMAP:
                    self._txi.features.isbumpmap = bool(int(args))
                    self._empty = False
                elif command == TXICommand.ISLIGHTMAP:
                    self._txi.features.islightmap = bool(int(args))
                    self._empty = False
                elif command == TXICommand.ISSPECULARBUMPMAP:
                    self._txi.features.isspecularbumpmap = bool(int(args))
                    self._empty = False
                elif command == TXICommand.LOWERRIGHTCOORDS:
                    self._txi.features.lowerrightcoords = [(0.0, 0.0, 0) for _ in range(int(args))]
                    mode = TXIReaderMode.LOWER_RIGHT_COORDS
                    cur_coords = 0
                    self._empty = False
                elif command == TXICommand.MAXSIZEHQ:
                    self._txi.features.maxSizeHQ = int(args)
                    self._empty = False
                elif command == TXICommand.MAXSIZELQ:
                    self._txi.features.maxSizeLQ = int(args)
                    self._empty = False
                elif command == TXICommand.MINSIZEHQ:
                    self._txi.features.minSizeHQ = int(args)
                    self._empty = False
                elif command == TXICommand.MINSIZELQ:
                    self._txi.features.minSizeLQ = int(args)
                    self._empty = False
                elif command == TXICommand.MIPMAP:
                    self._txi.features.mipmap = bool(int(args))
                    self._empty = False
                elif command == TXICommand.NUMCHARS:
                    self._txi.features.numchars = int(args)
                    self._empty = False
                elif command == TXICommand.NUMCHARSPERSHEET:
                    self._txi.features.numcharspersheet = int(args)
                    self._empty = False
                elif command == TXICommand.NUMX:
                    self._txi.features.numx = int(args)
                    self._empty = False
                elif command == TXICommand.NUMY:
                    self._txi.features.numy = int(args)
                    self._empty = False
                elif command == TXICommand.ONDEMAND:
                    self._txi.features.ondemand = bool(int(args))
                    self._empty = False
                elif command == TXICommand.PRIORITY:
                    self._txi.features.priority = int(args)
                    self._empty = False
                elif command == TXICommand.PROCEDURETYPE:
                    self._txi.features.proceduretype = args
                    self._empty = False
                elif command == TXICommand.ROWS:
                    self._txi.features.rows = int(args)
                    self._empty = False
                elif command == TXICommand.SPACINGB:
                    self._txi.features.spacingB = float(args)
                    self._empty = False
                elif command == TXICommand.SPACINGR:
                    self._txi.features.spacingR = float(args)
                    self._empty = False
                elif command == TXICommand.SPEED:
                    self._txi.features.speed = float(args)
                    self._empty = False
                elif command == TXICommand.TEMPORARY:
                    self._txi.features.temporary = bool(int(args))
                    self._empty = False
                elif command == TXICommand.TEXTUREWIDTH:
                    self._txi.features.texturewidth = float(args)
                    self._empty = False
                elif command == TXICommand.UNIQUE:
                    self._txi.features.unique = bool(int(args))
                    self._empty = False
                elif command == TXICommand.UPPERLEFTCOORDS:
                    self._txi.features.upperleftcoords = [(0.0, 0.0, 0) for _ in range(int(args))]
                    mode = TXIReaderMode.UPPER_LEFT_COORDS
                    cur_coords = 0
                    self._empty = False
                elif command == TXICommand.WATERHEIGHT:
                    self._txi.features.waterheight = float(args)
                    self._empty = False
                elif command == TXICommand.WATERWIDTH:
                    self._txi.features.waterwidth = float(args)
                    self._empty = False
                elif command == TXICommand.XBOX_DOWNSAMPLE:
                    self._txi.features.xbox_downsample = bool(int(args))
                    self._empty = False
            except Exception as e:  # noqa: BLE001
                RobustLogger().warning(f"Invalid TXI line: '{line}'", exc_info=e)

        return self._txi


class TXIBinaryWriter(ResourceWriter):
    def __init__(self, txi: TXI, target: TARGET_TYPES):
        super().__init__(target)
        self._txi: TXI = txi

    @autoclose
    def write(self, *, auto_close: bool = True):  # noqa: FBT001, FBT002, ARG002  # pyright: ignore[reportUnusedParameters]
        from pykotor.resource.formats.txi.txi_data import TXICommand

        lines: list[str] = []
        for attr, value in vars(self._txi.features).items():
            if value is None:
                continue
            if isinstance(value, list) and len(value) == 0:
                continue
            if isinstance(value, str) and len(value) == 0:
                continue
            upper_attr = attr.upper()
            if upper_attr not in TXICommand.__members__:
                RobustLogger().error(f"Invalid TXI attribute '{attr}'")
                continue
            command = TXICommand[upper_attr]
            if isinstance(value, bool):
                lines.append(f"{command.value} {int(value)}")
            elif isinstance(value, (int, float)):
                lines.append(f"{command.value} {value}")
            elif isinstance(value, list):
                if attr in [
                    TXICommand.UPPERLEFTCOORDS.value.lower(),
                    TXICommand.LOWERRIGHTCOORDS.value.lower(),
                ]:
                    lines.append(f"{command.value} {len(value)}")
                    lines.extend(" ".join(map(str, coord)) for coord in value)
                else:
                    lines.append(f"{command.value} {' '.join(map(str, value))}")
            else:
                lines.append(f"{command.value} {value}")
        txi_string = "\n".join(lines)
        self._writer.write_string(txi_string)



#!/usr/bin/env python3
from __future__ import annotations

import hashlib
import logging
import os
import platform
import stat
import sys

from collections import OrderedDict
from contextlib import suppress
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, SupportsFloat, SupportsInt, TypeVar

if TYPE_CHECKING:
    from collections.abc import Iterable
    from xml.etree.ElementTree import Element

    from typing_extensions import Buffer, Literal, Self, SupportsIndex

from utility.string_util import is_non_empty_string

T = TypeVar("T")
U = TypeVar("U")


class ProcessorArchitecture(Enum):
    BIT_32 = "32bit"
    BIT_64 = "64bit"

    def __str__(self) -> Literal["32bit", "64bit"]:
        return self.value

    def __int__(self) -> Literal[32, 64]:
        return self.get_int()

    @classmethod
    def from_os(cls) -> Self:
        return cls(platform.architecture()[0])

    @classmethod
    def from_python(cls) -> Literal[ProcessorArchitecture.BIT_64, ProcessorArchitecture.BIT_32]:
        return cls.BIT_64 if sys.maxsize > 2**32 else cls.BIT_32

    def get_machine_repr(self):
        return self._get("x86", "x64")

    def get_int(self) -> Literal[32, 64]:
        return self._get(32, 64)

    def get_dashed_bitness(self):
        return self._get("32-bit", "64-bit")

    def _get(self, arg0: T, arg1: U) -> T | U:  # sourcery skip: assign-if-exp, reintroduce-else  # noqa: ANN001
        if self == self.BIT_32:
            return arg0
        if self == self.BIT_64:
            return arg1
        raise RuntimeError(arg0, arg1)

    def supports_64_bit(self) -> bool:
        """Check if the architecture supports 64-bit processing."""
        return self == self.BIT_64


def format_gpu_info(
    info: Iterable,
    headers: tuple[str, ...],
) -> str:
    # Determine the maximum width for each column
    column_widths: list[int] = [max(len(str(row[i])) for row in (headers, *info)) for i in range(len(headers))]

    # Function to format a single row
    def format_row(row: Iterable) -> str:
        return " | ".join(f"{str(item).ljust(column_widths[i])}" for i, item in enumerate(row))

    # Build the output string
    output: str = format_row(headers) + "\n"
    output += "-" * (sum(column_widths) + len(headers) * 3 - 1) + "\n"  # Add a separator line

    # Append each row of GPU details to the output string
    for row in info:
        output += format_row(row) + "\n"

    return output


def print_excluding_base_classes(
    obj: object,
    exclude_base_classes: list[type[object]] | None = None,
):
    if exclude_base_classes is None:
        exclude_base_classes = [object, type]

    def get_base_class_attributes(base_classes: Iterable[type[object]]) -> set[str]:
        attrs = set()
        for base_class in base_classes:
            attrs.update(dir(base_class))
        return attrs

    def print_filtered_attributes(obj: object, obj_name: str, exclude_attrs: Iterable[str]):
        print(f"{obj_name} Attributes:")
        for attr in dir(obj):
            if not attr.startswith("_") and not callable(getattr(obj, attr)) and attr not in exclude_attrs:
                try:
                    print(f"  {attr}: {getattr(obj, attr)}")
                except Exception as ex:  # noqa: BLE001
                    print(f"  {attr}: Unable to retrieve value ({ex})")

    # Get the attributes of the base classes to exclude
    base_class_attrs = get_base_class_attributes(exclude_base_classes)

    # Print the filtered attributes of the object
    print_filtered_attributes(obj, obj.__class__.__name__, base_class_attrs)


def get_system_info() -> dict[str, Any]:
    # sourcery skip: extract-method, list-comprehension, merge-dict-assign
    info: dict[str, Any] = {}

    # Basic OS information
    info["Platform"] = platform.system()
    info["Platform-Release"] = platform.release()
    info["Platform-Version"] = platform.version()
    info["Architecture"] = platform.machine()

    # CPU information
    psutil = None
    with suppress(ImportError):
        import psutil  # pyright: ignore[reportMissingImports]  # type: ignore[no-redef]
    if psutil is not None:
        info["Physical cores"] = psutil.cpu_count(logical=False)
        info["Total cores"] = psutil.cpu_count(logical=True)
        cpu_freq = psutil.cpu_freq()
        info["Max Frequency"] = f"{cpu_freq.max:.2f}Mhz"
        info["Min Frequency"] = f"{cpu_freq.min:.2f}Mhz"
        info["Current Frequency"] = f"{cpu_freq.current:.2f}Mhz"
        info["CPU Usage Per Core"] = [f"{x}%" for x in psutil.cpu_percent(percpu=True, interval=1)]
        info["Total CPU Usage"] = f"{psutil.cpu_percent()}%"

        # RAM Information
        svmem = psutil.virtual_memory()
        info["Total Memory"] = f"{svmem.total / (1024**3):.2f} GB"
        info["Available Memory"] = f"{svmem.available / (1024**3):.2f} GB"
        info["Used Memory"] = f"{svmem.used / (1024**3):.2f} GB"
        info["Memory Usage"] = f"{svmem.percent}%"

    # GPU Information
    try:
        import GPUtil  # pyright: ignore[reportMissingImports]  # type: ignore[no-redef]
    except ImportError:
        if not TYPE_CHECKING:
            GPUtil = None
    if GPUtil is not None:
        gpus = GPUtil.getGPUs()
        gpu_info = []
        gpu_info.extend(
            (
                gpu.id,
                gpu.name,
                f"{gpu.memoryTotal}MB",
                f"{gpu.memoryUsed}MB",
                f"{gpu.memoryFree}MB",
                f"{gpu.driver}",
                f"{gpu.temperature} C",
            )
            for gpu in gpus
        )
        info["GPU Details"] = format_gpu_info(gpu_info, headers=("id", "name", "total memory", "used memory", "free memory", "driver", "temperature"))

    return info


def is_debug_mode() -> bool:
    ret = False
    if os.getenv("PYTHONDEBUG", None):
        ret = True
    if getattr(sys, "gettrace", None) is not None:
        ret = True
    if getattr(sys, "frozen", False) or getattr(sys, "_MEIPASS", False):
        ret = False
    if os.getenv("DEBUG_MODE", "0") == "1":
        ret = True
    return ret


def is_frozen() -> bool:
    return (
        getattr(sys, "frozen", False) or getattr(sys, "_MEIPASS", False)
        # or tempfile.gettempdir() in sys.executable
    )


def has_attr_excluding_object(
    cls: type,
    attr_name: str,
) -> bool:
    # Exclude the built-in 'object' class
    mro_classes: list[type] = [c for c in cls.mro() if c is not object]

    return any(attr_name in base_class.__dict__ for base_class in mro_classes)


def is_class_or_subclass_but_not_instance(
    cls: type,
    target_cls: type,
) -> bool:
    if cls is target_cls:
        return True
    if not hasattr(cls, "__bases__"):
        return False
    return any(is_class_or_subclass_but_not_instance(base, target_cls) for base in cls.__bases__)


def is_instance_or_subinstance(
    instance: object,
    target_cls: type,
) -> bool:
    if hasattr(instance, "__bases__"):  # instance is a class
        return False  # if instance is a class type, always return False
    # instance is not a class
    instance_type = instance.__class__
    return instance_type is target_cls or is_class_or_subclass_but_not_instance(instance_type, target_cls)


def generate_hash(
    data_input: bytes | bytearray | memoryview | os.PathLike | str,
    hash_algo: str = "sha1",  # sha1 is faster than md5 in python somehow
    chunk_size: int = 262144,  # 256KB default
    *,
    always_chunk: bool = False,  # Don't unnecessarily chunk bytes/bytearray inputs.
) -> str:
    try:
        hasher = hashlib.new(hash_algo)
    except ValueError as e:
        available_algos = ", ".join(sorted(hashlib.algorithms_available))
        msg = f"Invalid hash algorithm. Available algorithms are: [{available_algos}]"
        raise ValueError(msg) from e

    if isinstance(data_input, (bytes, bytearray, memoryview)):
        if always_chunk or isinstance(data_input, memoryview):
            for start in range(0, len(data_input), chunk_size):
                end = start + chunk_size
                hasher.update(data_input[start:end])
        else:
            hasher.update(data_input)
    else:
        with Path(data_input).open("rb") as f:
            for chunk in iter(lambda: f.read(chunk_size), b""):
                hasher.update(chunk)

    return hasher.hexdigest(64) if "shake" in hash_algo else hasher.hexdigest()  # type: ignore[call-arg]


def get_file_attributes(
    path: Path,
) -> dict[str, bool]:
    attributes: dict[str, bool] = {
        "is_hidden": False,
        "is_system": False,
        "is_archive": False,
        "is_compressed": False,
        "is_encrypted": False,
        "is_readonly": False,
        "is_temporary": False,
    }

    # Check if file is hidden
    attributes["is_hidden"] = (
        path.name.startswith(".") or bool(path.stat().st_file_attributes & stat.FILE_ATTRIBUTE_HIDDEN) if hasattr(stat, "FILE_ATTRIBUTE_HIDDEN") else False
    )

    # Check if file is read-only
    attributes["is_readonly"] = not os.access(path, os.W_OK)

    # Check if file is temporary (based on name)
    attributes["is_temporary"] = path.name.endswith(".tmp")

    # Platform-specific checks
    if os.name == "nt":  # Windows
        try:
            import ctypes

            FILE_ATTRIBUTE_SYSTEM = 0x4
            FILE_ATTRIBUTE_ARCHIVE = 0x20
            FILE_ATTRIBUTE_COMPRESSED = 0x800
            FILE_ATTRIBUTE_ENCRYPTED = 0x4000

            attrs: int = ctypes.windll.kernel32.GetFileAttributesW(str(path))
            if attrs != -1:
                attributes["is_system"] = bool(attrs & FILE_ATTRIBUTE_SYSTEM)
                attributes["is_archive"] = bool(attrs & FILE_ATTRIBUTE_ARCHIVE)
                attributes["is_compressed"] = bool(attrs & FILE_ATTRIBUTE_COMPRESSED)
                attributes["is_encrypted"] = bool(attrs & FILE_ATTRIBUTE_ENCRYPTED)
        except Exception:
            logging.getLogger(__name__).exception(f"Failed to get file attributes for: '{path}'")
    else:  # Unix-like systems
        # Check if file is system (based on location)
        attributes["is_system"] = str(path).startswith(("/etc", "/var", "/bin", "/sbin", "/usr/bin", "/usr/sbin"))

        # Check if file is archived (based on extension)
        archive_extensions = {".tar", ".gz", ".bz2", ".xz", ".zip", ".7z", ".rar"}
        attributes["is_archive"] = get_normalized_extension(path) in archive_extensions

        # Check if file is compressed (based on extension)
        compressed_extensions = {".gz", ".bz2", ".xz", ".zip", ".7z", ".rar"}
        attributes["is_compressed"] = get_normalized_extension(path) in compressed_extensions

        # Check if file is encrypted (based on extension, not reliable)
        encrypted_extensions = {".gpg", ".enc", ".asc"}
        attributes["is_encrypted"] = path.suffix.lower() in encrypted_extensions

    return attributes


def indent(
    elem: Element,
    level: int = 0,
):
    """Indents the XML element by the given level
    Args:
        elem: Element - The element to indent
        level: int - The level of indentation (default: 0).

    Returns:
    -------
        None - Indents the element in-place

    Processing Logic:
    ----------------
        - Calculate indentation string based on level
        - If element is empty, set text to indentation
        - If no tail, set tail to newline + indentation
        - Recursively indent child elements with increased level
        - If no tail after children, set tail to indentation
        - If level and no tail, set tail to indentation.
    """
    i: str = "\n" + level * "  "
    if len(elem):
        if not is_non_empty_string(elem.text):
            elem.text = f"{i}  "
        if not is_non_empty_string(elem.tail):
            elem.tail = i
        for e in elem:
            indent(e, level + 1)
        if not is_non_empty_string(elem.tail):
            elem.tail = i
    elif level and not is_non_empty_string(elem.tail):
        elem.tail = i


def is_int(
    val: str | int | Buffer | SupportsInt | SupportsIndex,
) -> bool:
    """Can be cast to an int without raising an error.

    Args:
    ----
        val (ConvertableToInt): The value to try to convert

    Returns:
    -------
        True if val can be converted else False
    """
    try:
        _ = int(val)
    except ValueError:
        return False
    else:
        return True


def is_float(
    val: str | float | Buffer | SupportsFloat | SupportsIndex,
) -> bool:
    """Can be cast to a float without raising an error.

    Args:
    ----
        val (ConvertableToFloat): The value to try to convert

    Returns:
    -------
        True if val can be converted else False
    """
    try:
        _ = float(val)
    except ValueError:
        return False
    else:
        return True


def to_kwargs(
    *args: Any,
    **kwargs: Any,
) -> dict[str, Any]:
    if sys.version_info < (3, 7):  # noqa: UP036
        kwargs = OrderedDict(kwargs)
    keys = iter(kwargs.keys())
    for arg in args:
        try:
            key = next(keys)
            if kwargs[key] is None:
                kwargs[key] = arg
        except StopIteration as e:  # noqa: PERF203
            raise ValueError("Too many positional arguments for the available keyword arguments.") from e  # noqa: B904
    return dict(kwargs)


def is_valid_path(path: Path | str | None) -> bool:
    """Check if a path is valid (not None and exists).

    Args:
    ----
        path: The path to validate

    Returns:
    -------
        True if path is not None and exists, False otherwise
    """
    return path is not None and Path(path).exists()


def get_normalized_extension(path: Path | str) -> str:
    """Get the file extension in lowercase.

    Args:
    ----
        path: The file path

    Returns:
    -------
        The file extension in lowercase (e.g., '.txt', '.mdl')
    """
    return Path(path).suffix.lower()


def ensure_directory_exists(path: Path | str) -> None:
    """Ensure that a directory exists, creating it if necessary.

    Args:
    ----
        path: The directory path to ensure exists
    """
    Path(path).mkdir(parents=True, exist_ok=True)


def format_exception_message(exception: Exception) -> str:
    """Format an exception into a consistent string format.

    Args:
    ----
        exception: The exception to format

    Returns:
    -------
        Formatted exception message as "ExceptionClass: message"
    """
    return f"{exception.__class__.__name__}: {exception}"


def is_installation_path(path: object) -> bool:
    """Check if a path object is an Installation instance.

    Args:
    ----
        path: The path object to check

    Returns:
    -------
        True if the path is an Installation instance, False otherwise
    """
    from pykotor.extract.installation import Installation  # Import here to avoid circular imports
    return isinstance(path, Installation)
